# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque
from typing import Dict, List, Optional, Type

import onnx

from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.operator_metatypes import OperatorMetatypeRegistry
from nncf.common.hardware.opset import HWConfigOpName
from nncf.onnx.graph.onnx_helper import get_parent
from nncf.onnx.graph.onnx_helper import get_parents_node_mapping
from nncf.onnx.graph.onnx_helper import get_tensor
from nncf.onnx.graph.onnx_helper import has_tensor

ONNX_OPERATION_METATYPES = OperatorMetatypeRegistry("onnx_operator_metatypes")


class ONNXOpMetatype(OperatorMetatype):
    op_names: List[str] = []
    subtypes: List[Type[OperatorMetatype]] = []

    @classmethod
    def get_all_aliases(cls) -> List[str]:
        return cls.op_names

    @classmethod
    def get_subtypes(cls) -> List[Type[OperatorMetatype]]:
        return cls.subtypes

    @classmethod
    def matches(cls, model: onnx.ModelProto, node: onnx.NodeProto) -> bool:
        return node.op_type in cls.op_names

    @classmethod
    def determine_subtype(cls, model: onnx.ModelProto, node: onnx.NodeProto) -> Optional[Type[OperatorMetatype]]:
        matches = []
        subtypes_list = deque(cls.get_subtypes())
        while subtypes_list:
            subtype = subtypes_list.popleft()
            if subtype.matches(model, node):
                subtypes_list.extend(subtype.get_subtypes())
                matches.append(subtype)
        if not matches:
            return None
        return matches[-1]


class ONNXOpWithWeightsMetatype(ONNXOpMetatype):
    """
    Metatype which could have weights.
    :param weight_channel_axis: Axis for weight per-channel quantization.
    :param weight_port_ids: Constant input ports of the node's weight. Defaults to an empty list.
    :param bias_port_id: Input port of the node's bias. If the value is None,
    it means that the Metatype does not have bias. Defaults to None.
    :param possible_weight_ports: Input ports on which weight could be laid. Defaults to an empty list.
    """

    weight_channel_axis: int
    weight_port_ids: List[int] = []
    bias_port_id: Optional[int] = None
    possible_weight_ports: List[int] = []


@ONNX_OPERATION_METATYPES.register(is_subtype=True)
class ONNXDepthwiseConvolutionMetatype(ONNXOpWithWeightsMetatype):
    name = "DepthwiseConvOp"
    op_names = ["Conv"]
    hw_config_names = [HWConfigOpName.DEPTHWISECONVOLUTION]
    weight_channel_axis = 0
    weight_port_ids = [1]
    bias_port_id = 2
    output_channel_axis = 1

    @classmethod
    def matches(cls, model: onnx.ModelProto, node: onnx.NodeProto) -> bool:
        return _is_depthwise_conv(model, node)


@ONNX_OPERATION_METATYPES.register(is_subtype=True)
class ONNXGroupConvolutionMetatype(ONNXOpWithWeightsMetatype):
    name = "GroupConvOp"
    op_names = ["Conv"]
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    weight_channel_axis = 0
    weight_port_ids = [1]
    bias_port_id = 2
    output_channel_axis = 1
    subtypes = [ONNXDepthwiseConvolutionMetatype]

    @classmethod
    def matches(cls, model: onnx.ModelProto, node: onnx.NodeProto) -> bool:
        return _is_group_conv(node)


@ONNX_OPERATION_METATYPES.register()
class ONNXConvolutionMetatype(ONNXOpWithWeightsMetatype):
    name = "ConvOp"
    op_names = ["Conv"]
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    weight_channel_axis = 0
    weight_port_ids = [1]
    bias_port_id = 2
    output_channel_axis = 1
    subtypes = [ONNXGroupConvolutionMetatype]


@ONNX_OPERATION_METATYPES.register()
class ONNXConvolutionTransposeMetatype(ONNXOpWithWeightsMetatype):
    name = "ConvTransposeOp"
    op_names = ["ConvTranspose"]
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    weight_channel_axis = 1
    weight_port_ids = [1]
    bias_port_id = 2
    output_channel_axis = 1


@ONNX_OPERATION_METATYPES.register()
class ONNXGemmMetatype(ONNXOpWithWeightsMetatype):
    name = "GemmOp"
    op_names = ["Gemm"]
    hw_config_names = [HWConfigOpName.MATMUL]
    weight_channel_axis = -1  # For port_id=1
    bias_port_id = 2
    possible_weight_ports = [0, 1]
    output_channel_axis = -1


@ONNX_OPERATION_METATYPES.register()
class ONNXMatMulMetatype(ONNXOpWithWeightsMetatype):
    name = "MatMulOp"
    op_names = ["MatMul"]
    hw_config_names = [HWConfigOpName.MATMUL]
    weight_channel_axis = -1  # For port_id=1
    bias_port_id = 2
    possible_weight_ports = [0, 1]
    output_channel_axis = -1


@ONNX_OPERATION_METATYPES.register()
class ONNXReluMetatype(ONNXOpMetatype):
    name = "ReluOp"
    op_names = ["Relu", "Clip"]


@ONNX_OPERATION_METATYPES.register()
class ONNXLeakyReluMetatype(ONNXOpMetatype):
    name = "LeakyReluOp"
    op_names = ["LeakyRelu"]


@ONNX_OPERATION_METATYPES.register()
class ONNXThresholdedReluMetatype(ONNXOpMetatype):
    name = "ThresholdedReluOp"
    op_names = ["ThresholdedRelu"]


@ONNX_OPERATION_METATYPES.register()
class ONNXEluMetatype(ONNXOpMetatype):
    name = "EluOp"
    op_names = ["Elu"]


@ONNX_OPERATION_METATYPES.register()
class ONNXPReluMetatype(ONNXOpMetatype):
    name = "PReluOp"
    op_names = ["PRelu"]


@ONNX_OPERATION_METATYPES.register()
class ONNXSigmoidMetatype(ONNXOpMetatype):
    name = "SigmoidOp"
    op_names = ["Sigmoid"]


@ONNX_OPERATION_METATYPES.register()
class ONNXHardSigmoidMetatype(ONNXOpMetatype):
    name = "HardSigmoidOp"
    op_names = ["HardSigmoid"]


@ONNX_OPERATION_METATYPES.register()
class ONNXHardSwishMetatype(ONNXOpMetatype):
    name = "HardSwishOp"
    op_names = ["HardSwish"]


@ONNX_OPERATION_METATYPES.register()
class ONNXGlobalAveragePoolMetatype(ONNXOpMetatype):
    name = "GlobalAveragePoolOp"
    op_names = ["GlobalAveragePool"]
    hw_config_names = [HWConfigOpName.AVGPOOL]


@ONNX_OPERATION_METATYPES.register()
class ONNXAveragePoolMetatype(ONNXOpMetatype):
    name = "AveragePoolOp"
    op_names = ["AveragePool"]
    hw_config_names = [HWConfigOpName.AVGPOOL]


@ONNX_OPERATION_METATYPES.register()
class ONNXGlobalMaxPoolMetatype(ONNXOpMetatype):
    name = "GlobalMaxPoolOp"
    op_names = ["GlobalMaxPool"]
    hw_config_names = [HWConfigOpName.MAXPOOL]


@ONNX_OPERATION_METATYPES.register()
class ONNXMaxPoolMetatype(ONNXOpMetatype):
    name = "MaxPoolOp"
    op_names = ["MaxPool"]
    hw_config_names = [HWConfigOpName.MAXPOOL]


@ONNX_OPERATION_METATYPES.register()
class ONNXConstantMetatype(ONNXOpMetatype):
    name = "ConstantOp"
    op_names = ["Constant"]


@ONNX_OPERATION_METATYPES.register()
class ONNXAddLayerMetatype(ONNXOpMetatype):
    name = "AddOp"
    op_names = ["Add", "Sum"]
    hw_config_names = [HWConfigOpName.ADD]


@ONNX_OPERATION_METATYPES.register()
class ONNXSubMetatype(ONNXOpMetatype):
    name = "SubOp"
    op_names = ["Sub"]
    hw_config_names = [HWConfigOpName.SUBTRACT]


@ONNX_OPERATION_METATYPES.register()
class ONNXMulLayerMetatype(ONNXOpMetatype):
    name = "MulOp"
    op_names = ["Mul"]
    hw_config_names = [HWConfigOpName.MULTIPLY]


@ONNX_OPERATION_METATYPES.register()
class ONNXDivLayerMetatype(ONNXOpMetatype):
    name = "DivOp"
    op_names = ["Div"]
    hw_config_names = [HWConfigOpName.DIVIDE]


@ONNX_OPERATION_METATYPES.register()
class ONNXConcatMetatype(ONNXOpMetatype):
    name = "ConcatOp"
    op_names = ["Concat"]
    hw_config_names = [HWConfigOpName.CONCAT]


@ONNX_OPERATION_METATYPES.register()
class ONNXBatchNormMetatype(ONNXOpMetatype):
    name = "BatchNormalizationOp"
    op_names = ["BatchNormalization"]


@ONNX_OPERATION_METATYPES.register()
class ONNXResizeMetatype(ONNXOpMetatype):
    name = "ResizeOp"
    op_names = ["Resize"]
    hw_config_names = [HWConfigOpName.INTERPOLATE]


@ONNX_OPERATION_METATYPES.register()
class ONNXCenterCropPadMetatype(ONNXOpMetatype):
    name = "CenterCropPadOp"
    op_names = ["CenterCropPad"]
    hw_config_names = [HWConfigOpName.CROP]


@ONNX_OPERATION_METATYPES.register()
class ONNXReshapeMetatype(ONNXOpMetatype):
    name = "ReshapeOp"
    op_names = ["Reshape"]
    hw_config_names = [HWConfigOpName.RESHAPE]


@ONNX_OPERATION_METATYPES.register()
class ONNXTileMetatype(ONNXOpMetatype):
    name = "TileOp"
    op_names = ["Tile"]


@ONNX_OPERATION_METATYPES.register()
class ONNXUpsampleMetatype(ONNXOpMetatype):
    name = "UpsampleOp"
    op_names = ["Upsample"]


@ONNX_OPERATION_METATYPES.register()
class ONNXConstantOfShapeMetatype(ONNXOpMetatype):
    name = "ConstantOfShapeOp"
    op_names = ["ConstantOfShape"]


@ONNX_OPERATION_METATYPES.register()
class ONNXShapeMetatype(ONNXOpMetatype):
    name = "ShapeOp"
    op_names = ["Shape"]


@ONNX_OPERATION_METATYPES.register()
class ONNXExpandMetatype(ONNXOpMetatype):
    name = "ExpandOp"
    op_names = ["Expand"]


@ONNX_OPERATION_METATYPES.register()
class ONNXNonZeroMetatype(ONNXOpMetatype):
    name = "NonZeroOp"
    op_names = ["NonZero"]


@ONNX_OPERATION_METATYPES.register()
class ONNXSplitMetatype(ONNXOpMetatype):
    name = "SplitOp"
    op_names = ["Split"]
    hw_config_names = [HWConfigOpName.SPLIT]


@ONNX_OPERATION_METATYPES.register()
class ONNXLessMetatype(ONNXOpMetatype):
    name = "LessOp"
    op_names = ["Less"]
    hw_config_names = [HWConfigOpName.LESS]


@ONNX_OPERATION_METATYPES.register()
class ONNXLessOrEqualMetatype(ONNXOpMetatype):
    name = "LessOrEqualOp"
    op_names = ["LessOrEqual"]
    hw_config_names = [HWConfigOpName.LESSEQUAL]


@ONNX_OPERATION_METATYPES.register()
class ONNXGreaterMetatype(ONNXOpMetatype):
    name = "GreaterOp"
    op_names = ["Greater"]
    hw_config_names = [HWConfigOpName.GREATER]


@ONNX_OPERATION_METATYPES.register()
class ONNXGreaterOrEqualMetatype(ONNXOpMetatype):
    name = "GreaterOrEqualOp"
    op_names = ["GreaterOrEqual"]
    hw_config_names = [HWConfigOpName.GREATEREQUAL]


@ONNX_OPERATION_METATYPES.register()
class ONNXEqualMetatype(ONNXOpMetatype):
    name = "EqualOp"
    op_names = ["Equal"]
    hw_config_names = [HWConfigOpName.EQUAL]


@ONNX_OPERATION_METATYPES.register()
class ONNXNotMetatype(ONNXOpMetatype):
    name = "NotOp"
    op_names = ["Not"]
    hw_config_names = [HWConfigOpName.LOGICALNOT]


@ONNX_OPERATION_METATYPES.register()
class ONNXAndMetatype(ONNXOpMetatype):
    name = "AndOp"
    op_names = ["And"]
    hw_config_names = [HWConfigOpName.LOGICALAND]


@ONNX_OPERATION_METATYPES.register()
class ONNXOrMetatype(ONNXOpMetatype):
    name = "OrOp"
    op_names = ["Or"]
    hw_config_names = [HWConfigOpName.LOGICALOR]


@ONNX_OPERATION_METATYPES.register()
class ONNXXOrMetatype(ONNXOpMetatype):
    name = "XorOp"
    op_names = ["Xor"]
    hw_config_names = [HWConfigOpName.LOGICALXOR]


@ONNX_OPERATION_METATYPES.register()
class ONNXModMetatype(ONNXOpMetatype):
    name = "ModOp"
    op_names = ["Mod"]
    hw_config_names = [HWConfigOpName.FLOORMOD]


@ONNX_OPERATION_METATYPES.register()
class ONNXMaximumMetatype(ONNXOpMetatype):
    name = "MaxOp"
    op_names = ["Max"]
    hw_config_names = [HWConfigOpName.MAXIMUM]


@ONNX_OPERATION_METATYPES.register()
class ONNXMinimumMetatype(ONNXOpMetatype):
    name = "MinOp"
    op_names = ["Min"]
    hw_config_names = [HWConfigOpName.MINIMUM]


@ONNX_OPERATION_METATYPES.register()
class ONNXMeanMetatype(ONNXOpMetatype):
    name = "MeanOp"
    op_names = ["Mean"]


@ONNX_OPERATION_METATYPES.register()
class ONNXFloorMetatype(ONNXOpMetatype):
    name = "FloorOp"
    op_names = ["Floor"]


@ONNX_OPERATION_METATYPES.register()
class ONNXPowMetatype(ONNXOpMetatype):
    name = "PowOp"
    op_names = ["Pow"]
    hw_config_names = [HWConfigOpName.POWER]


@ONNX_OPERATION_METATYPES.register()
class ONNXSqrtMetatype(ONNXOpMetatype):
    name = "SqrtOp"
    op_names = ["Sqrt"]
    hw_config_names = [HWConfigOpName.POWER]


@ONNX_OPERATION_METATYPES.register()
class ONNXReciprocalMetatype(ONNXOpMetatype):
    name = "ReciprocalOp"
    op_names = ["Reciprocal"]
    hw_config_names = [HWConfigOpName.POWER]


@ONNX_OPERATION_METATYPES.register(is_subtype=True)
class ONNXEmbeddingMetatype(ONNXOpWithWeightsMetatype):
    name = "EmbeddingOp"
    hw_config_names = [HWConfigOpName.EMBEDDING]
    weight_port_ids = [0]
    weight_channel_axis = 0

    @classmethod
    def matches(cls, model: onnx.ModelProto, node: onnx.NodeProto) -> bool:
        return _is_embedding(model, node)


@ONNX_OPERATION_METATYPES.register()
class ONNXLogMetatype(ONNXOpMetatype):
    name = "LogOp"
    op_names = ["Log"]


@ONNX_OPERATION_METATYPES.register()
class ONNXAbsMetatype(ONNXOpMetatype):
    name = "AbsOp"
    op_names = ["Abs"]


@ONNX_OPERATION_METATYPES.register()
class ONNXScatterElementsMetatype(ONNXOpMetatype):
    name = "ScatterElementsOp"
    op_names = ["ScatterElements"]


@ONNX_OPERATION_METATYPES.register()
class ONNXScatterMetatype(ONNXOpMetatype):
    name = "ScatterOp"
    op_names = ["Scatter"]


@ONNX_OPERATION_METATYPES.register()
class ONNXScatterNDMetatype(ONNXOpMetatype):
    name = "ScatterNDOp"
    op_names = ["ScatterND"]


@ONNX_OPERATION_METATYPES.register()
class ONNXROIAlignMetatype(ONNXOpMetatype):
    name = "ROIAlignOp"
    op_names = ["RoiAlign"]


@ONNX_OPERATION_METATYPES.register()
class ONNXGatherMetatype(ONNXOpMetatype):
    name = "GatherOp"
    op_names = ["Gather"]
    subtypes = [ONNXEmbeddingMetatype]


@ONNX_OPERATION_METATYPES.register()
class ONNXGatherNDMetatype(ONNXOpMetatype):
    name = "GatherNDOp"
    op_names = ["GatherND"]


@ONNX_OPERATION_METATYPES.register()
class ONNXGatherElementsMetatype(ONNXOpMetatype):
    name = "GatherElementsOp"
    op_names = ["GatherElements"]


@ONNX_OPERATION_METATYPES.register()
class ONNXUnsqueezeMetatype(ONNXOpMetatype):
    name = "UnsqueezeOp"
    op_names = ["Unsqueeze"]
    hw_config_names = [HWConfigOpName.UNSQUEEZE]


@ONNX_OPERATION_METATYPES.register()
class ONNXSqueezeMetatype(ONNXOpMetatype):
    name = "SqueezeOp"
    op_names = ["Squeeze"]
    hw_config_names = [HWConfigOpName.SQUEEZE]


@ONNX_OPERATION_METATYPES.register()
class ONNXNonMaxSuppressionMetatype(ONNXOpMetatype):
    name = "NonMaxSuppressionOp"
    op_names = ["NonMaxSuppression"]


@ONNX_OPERATION_METATYPES.register()
class ONNXCastMetatype(ONNXOpMetatype):
    name = "CastOp"
    op_names = ["Cast"]


@ONNX_OPERATION_METATYPES.register()
class ONNXCastLikeMetatype(ONNXOpMetatype):
    name = "CastLikeOp"
    op_names = ["CastLike"]


@ONNX_OPERATION_METATYPES.register()
class ONNXReduceMinMetatype(ONNXOpMetatype):
    name = "ReduceMinOp"
    op_names = ["ReduceMin"]


@ONNX_OPERATION_METATYPES.register()
class ONNXReduceMaxMetatype(ONNXOpMetatype):
    name = "ReduceMaxOp"
    op_names = ["ReduceMax"]
    hw_config_names = [HWConfigOpName.REDUCEMAX]


@ONNX_OPERATION_METATYPES.register()
class ONNXReduceSumMetatype(ONNXOpMetatype):
    name = "ReduceSumOp"
    op_names = ["ReduceSum"]
    hw_config_names = [HWConfigOpName.REDUCESUM]


class ONNXReduceL2Metatype(ONNXOpMetatype):
    name = "ReduceL2Op"
    op_names = ["ReduceL2"]
    hw_config_names = [HWConfigOpName.REDUCEL2]


@ONNX_OPERATION_METATYPES.register()
class ONNXDepthToSpaceMetatype(ONNXOpMetatype):
    name = "DepthToSpaceOp"
    op_names = ["DepthToSpace"]


@ONNX_OPERATION_METATYPES.register()
class ONNXSpaceToDepthMetatype(ONNXOpMetatype):
    name = "SpaceToDepthOp"
    op_names = ["SpaceToDepth"]


@ONNX_OPERATION_METATYPES.register()
class ONNXReduceMeanMetatype(ONNXOpMetatype):
    name = "ReduceMeanOp"
    op_names = ["ReduceMean"]
    hw_config_names = [HWConfigOpName.REDUCEMEAN]


@ONNX_OPERATION_METATYPES.register()
class ONNXTopKMetatype(ONNXOpMetatype):
    name = "TopKOp"
    op_names = ["TopK"]


@ONNX_OPERATION_METATYPES.register()
class ONNXSliceMetatype(ONNXOpMetatype):
    name = "SliceOp"
    op_names = ["Slice"]


@ONNX_OPERATION_METATYPES.register()
class ONNXExpMetatype(ONNXOpMetatype):
    name = "ExpOp"
    op_names = ["Exp"]


@ONNX_OPERATION_METATYPES.register()
class ONNXTransposeMetatype(ONNXOpMetatype):
    name = "TransposeOp"
    op_names = ["Transpose"]
    hw_config_names = [HWConfigOpName.TRANSPOSE]


@ONNX_OPERATION_METATYPES.register()
class ONNXDropoutMetatype(ONNXOpMetatype):
    name = "DropoutOp"
    op_names = ["Dropout"]


@ONNX_OPERATION_METATYPES.register()
class ONNXFlattenMetatype(ONNXOpMetatype):
    name = "FlattenOp"
    op_names = ["Flatten"]
    hw_config_names = [HWConfigOpName.FLATTEN]


@ONNX_OPERATION_METATYPES.register()
class ONNXSoftmaxMetatype(ONNXOpMetatype):
    name = "SoftmaxOp"
    op_names = ["Softmax"]


@ONNX_OPERATION_METATYPES.register()
class ONNXPadMetatype(ONNXOpMetatype):
    name = "PadOp"
    op_names = ["Pad"]


@ONNX_OPERATION_METATYPES.register()
class ONNXIdentityMetatype(ONNXOpMetatype):
    name = "IdentityOp"
    op_names = ["Identity"]


@ONNX_OPERATION_METATYPES.register()
class ONNXQuantizeLinearMetatype(ONNXOpMetatype):
    name = "QuantizeLinearOp"
    op_names = ["QuantizeLinear"]


@ONNX_OPERATION_METATYPES.register()
class ONNXDequantizeLinearMetatype(ONNXOpMetatype):
    name = "DequantizeLinearOp"
    op_names = ["DequantizeLinear"]


@ONNX_OPERATION_METATYPES.register()
class ONNXDeformableConvolutionMetatype(ONNXOpMetatype):
    name = "DeformConvOp"
    op_names = ["DeformConv"]


@ONNX_OPERATION_METATYPES.register()
class ONNXErfMetatype(ONNXOpMetatype):
    name = "ErfOp"
    op_names = ["Erf"]


@ONNX_OPERATION_METATYPES.register()
class ONNXCosMetatype(ONNXOpMetatype):
    name = "CosOp"
    op_names = ["Cos"]


@ONNX_OPERATION_METATYPES.register()
class ONNXSinMetatype(ONNXOpMetatype):
    name = "SinOp"
    op_names = ["Sin"]


def get_operator_metatypes() -> List[Type[OperatorMetatype]]:
    """
    Returns a list of the operator metatypes.

    :return: List of operator metatypes.
    """
    return list(ONNX_OPERATION_METATYPES.registry_dict.values())


def get_metatype(model: onnx.ModelProto, node: onnx.NodeProto) -> ONNXOpMetatype:
    """
    Returns matched ONNXOpMetatype metatype to a ONNX node.

    :param model: ONNX model.
    :param node: Node from ONNX model.
    :return: Matched metatype.
    """
    metatype = ONNX_OPERATION_METATYPES.get_operator_metatype_by_op_name(node.op_type)
    if metatype.get_subtypes():
        subtype = metatype.determine_subtype(model, node)
        if subtype is not None:
            metatype = subtype
    return metatype


def get_tensor_edge_name(
    model: onnx.ModelProto,
    node: onnx.NodeProto,
    port_id: int,
    parents_node_mapping: Dict[str, onnx.NodeProto],
) -> Optional[str]:
    """
    Returns an edge name associated with a weight of a node laying on  an input port_id.

    Checks whether a node has a tensor on input port_id.
    If does then it is a weight and returns corresponding edge name.
    If not - take a parent node into this port id and does the same check for it.

    If an edge with a weight was not found then returns None.

    METATYPES THAT COULD CONSUME A WEIGHT TENSOR:
        ONNXConstantMetatype
        ONNXIdentityMetatype
        ONNXReshapeMetatype
        ONNXTransposeMetatype
        ONNXQuantizeLinearMetatype

    :param model: ONNX model.
    :param node: Node.
    :param port_id: Port id on which a weight edge is seeking.
    :param parents_node_mapping: Mapping from edge name to node which outputs this edge.
    :return: Edge name associated with a weight.
    """
    PROPAGATING_NODES = (
        ONNXIdentityMetatype.get_all_aliases()
        + ONNXTransposeMetatype.get_all_aliases()
        + ONNXQuantizeLinearMetatype.get_all_aliases()
        + ONNXReshapeMetatype.get_all_aliases()
        + ONNXDequantizeLinearMetatype.get_all_aliases()
    )
    END_NODES = ONNXConstantMetatype.get_all_aliases()
    parent = get_parent(node, port_id, parents_node_mapping)
    if not parent:
        if has_tensor(model, node.input[port_id]):
            return node.input[port_id]
    elif parent.op_type in END_NODES:
        return node.input[port_id]
    elif parent.op_type in PROPAGATING_NODES:
        return get_tensor_edge_name(model, parent, 0, parents_node_mapping)
    return None


def _is_group_conv(node: onnx.NodeProto) -> bool:
    """
    Returns True if the convolution is group, False - otherwise.
    Group convolution is a convolution with the group attribute.

    :param node: Convolution node to check whether it is depthwise.
    :return: True if the convolution is group, False - otherwise.
    """
    conv_group = None
    for attribute in node.attribute:
        if attribute.name == "group":
            conv_group = onnx.helper.get_attribute_value(attribute)
    if conv_group is None or conv_group == 1:
        return False
    return True


def _is_depthwise_conv(model: onnx.ModelProto, node: onnx.NodeProto) -> bool:
    """
    Returns True if the convolution is depthwise, False - otherwise.
    Depthwise convolution is a convolution satisfies the following rule:
    groups == in_channels and out_channels == K*in_channels, where K is a positive integer.
    Weight tensor of a convolution consists of the following dimension:
    (out_channels, in_channels / groups, kernel_size[0], kernel_size[1]).

    :param model: ONNX model to get the node's weight.
    :param node: Convolution node to check whether it is depthwise.
    :return: True if the convolution is depthwise, False - otherwise.
    """
    for attribute in node.attribute:
        if attribute.name == "group":
            conv_group = onnx.helper.get_attribute_value(attribute)
    weight_tensor_value = None
    initializer_name = get_tensor_edge_name(model, node, 1, get_parents_node_mapping(model))
    for init in model.graph.initializer:
        if init.name == initializer_name:
            weight_tensor_value = onnx.numpy_helper.to_array(init)
    if weight_tensor_value is None:
        return False
    conv_out_channels = weight_tensor_value.shape[0]
    conv_in_channels = weight_tensor_value.shape[1] * conv_group
    if (
        conv_out_channels % conv_in_channels == 0
        and conv_out_channels // conv_in_channels > 0
        and conv_group == conv_in_channels
    ):
        return True
    return False


def _is_embedding(model: onnx.ModelProto, node: onnx.NodeProto) -> bool:
    """
    Returns True if the layer can be represented as embedding, False - otherwise.

    :param model: ONNX model to get the node's weight.
    :param node: Layer to check whether it is embedding.
    :return: True if the layer is embedding, False - otherwise.
    """
    tensor_port_id = ONNXEmbeddingMetatype.weight_port_ids[0]
    allowed_types_list = ["TensorProto.FLOAT"]
    parents_node_mapping = get_parents_node_mapping(model)
    weight_edge_name = get_tensor_edge_name(model, node, tensor_port_id, parents_node_mapping)

    if weight_edge_name is not None:
        tensor_data_type = get_tensor(model, weight_edge_name).data_type
        if onnx.helper.tensor_dtype_to_string(tensor_data_type) in allowed_types_list:
            return True
    return False



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque
from copy import deepcopy
from itertools import chain
from typing import Any, Dict, List, Optional, Tuple

import networkx as nx
import numpy as np
import torch

from nncf.common.collector import StatisticsCollector
from nncf.common.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.graph_matching import find_subgraphs_matching_pattern
from nncf.common.graph.patterns.manager import PatternsManager
from nncf.common.graph.patterns.manager import TargetDevice
from nncf.common.quantization.collectors import QuantizationStatisticsCollector
from nncf.common.quantization.collectors import QuantizerDescription
from nncf.common.quantization.quantizer_propagation.structs import QuantizationTrait
from nncf.common.quantization.structs import NonWeightQuantizerId
from nncf.common.quantization.structs import WeightQuantizerId
from nncf.common.utils.backend import BackendType
from nncf.common.utils.debug import is_debug
from nncf.torch.nncf_module_replacement import is_nncf_module
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.nncf_network import PTNNCFGraph
from nncf.torch.quantization.default_quantization import DEFAULT_PT_QUANT_TRAIT_TO_OP_DICT
from nncf.torch.quantization.layers import BaseQuantizer
from nncf.torch.quantization.layers import SymmetricQuantizer
from nncf.torch.quantization.statistics import MemoryConsumptionStatistics
from nncf.torch.quantization.statistics import QuantizationConfigurationStatistics
from nncf.torch.quantization.structs import NonWeightQuantizerInfo
from nncf.torch.quantization.structs import WeightQuantizerInfo


class QuantizationShareBuildTimeInfo:
    def __init__(self, aq_potential_num: int, wq_potential_num: int):
        self.aq_potential_num = aq_potential_num
        self.wq_potential_num = wq_potential_num

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {"aq_potential_num": self.aq_potential_num, "wq_potential_num": self.wq_potential_num}

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "QuantizationShareBuildTimeInfo":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        return cls(**state)


class PTQuantizationStatisticsCollector(QuantizationStatisticsCollector):
    """
    Implementation of the quantization statistics collector for the PyTorch backend.
    """

    def __init__(
        self,
        weight_quantizers: Dict[WeightQuantizerId, WeightQuantizerInfo],
        non_weight_quantizers: Dict[NonWeightQuantizerId, NonWeightQuantizerInfo],
        build_time_info: QuantizationShareBuildTimeInfo,
    ):
        """
        Initializes a collector of the quantization statistics.
        """
        self._weight_quantizers = {k: v.quantizer_module_ref for k, v in weight_quantizers.items()}
        self._non_weight_quantizers = {k: v.quantizer_module_ref for k, v in non_weight_quantizers.items()}
        self._info = build_time_info

    def _collect_quantizers_descriptions(self) -> List[QuantizerDescription]:
        """
        Collects descriptions of the quantizers.

        :return: Descriptions of the quantizers.
        """
        # `True` for weight quantizer, `False` otherwise.
        quantizers = chain(
            map(lambda x: (True, x), self._weight_quantizers.values()),
            map(lambda x: (False, x), self._non_weight_quantizers.values()),
        )

        quantizers_descriptions = []
        for is_weight_quantizer, q in quantizers:
            is_symmetric = isinstance(q, SymmetricQuantizer)

            quantizers_descriptions.append(
                QuantizerDescription(
                    q.num_bits, q.per_channel, q.signed, is_symmetric, is_weight_quantizer, q.is_enabled_quantization()
                )
            )

        return quantizers_descriptions

    def _get_potential_quantizers_num(self) -> Tuple[int, int]:
        """
        Returns a potential number of quantizers for weights and activations.

        :return: A tuple (wq_potential_num, aq_potential_num) where
            - `wq_potential_num` is a potential number of quantizers for weights.
            - `aq_potential_num` is a potential number of quantizers for activations.
        """
        aq_potential_num = self._info.aq_potential_num if is_debug() else None
        return self._info.wq_potential_num, aq_potential_num


class MemoryConsumptionStatisticsCollector(StatisticsCollector):
    """
    This metric considers:
        - how many times memory consumption for network weights will decrease.
        - how many times memory consumption* for activations tensor will decrease.

    * Reflects host memory consumption, assuming only the final low-precision output activation tensors are stored
      in host memory (i.e. assuming intermediate accumulation results are only stored in device memory)
    """

    def __init__(
        self,
        compressed_model: NNCFNetwork,
        weight_quantizers: Dict[WeightQuantizerId, WeightQuantizerInfo],
        non_weight_quantizers: Dict[NonWeightQuantizerId, NonWeightQuantizerInfo],
    ):
        """
        Initializes collector of the memory consumption statistics.
        """
        self._compressed_model = compressed_model
        self._weight_quantizers = weight_quantizers
        self._non_weight_quantizers = non_weight_quantizers

    def collect(self) -> MemoryConsumptionStatistics:
        stats = MemoryConsumptionStatistics()

        fp_num_bits = 32
        nncf_modules = self._compressed_model.nncf.get_nncf_modules()
        for nncf_module in nncf_modules:
            count_el = np.prod(nncf_module.weight.shape)
            stats.fp32_weight_size += count_el * fp_num_bits
            quantizer = self._get_weight_quantizer_for_module(nncf_module)
            if quantizer is not None:
                num_bits = quantizer.num_bits
                stats.quantized_weight_size += count_el * num_bits
            else:
                stats.quantized_weight_size += count_el * fp_num_bits

        try:
            stats.weight_memory_consumption_decrease = stats.fp32_weight_size / stats.quantized_weight_size
        except ZeroDivisionError:
            stats.weight_memory_consumption_decrease = 0

        stats.quantized_weight_size /= 2**23
        stats.fp32_weight_size /= 2**23

        original_graph = deepcopy(self._compressed_model.nncf.get_original_graph())

        memory_consumption_fp_model = {}
        memory_consumption_compressed_model = {}

        original_nx_graph = original_graph._nx_graph
        nx.set_edge_attributes(original_nx_graph, 32, "precision")

        for u, v in original_nx_graph.edges:
            shape = original_nx_graph.edges[u, v][NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR]
            num_bits = self._get_precision_for_activation_tensor(u, v, original_nx_graph)
            original_nx_graph.edges[u, v]["precision"] = num_bits
            u_node_name = original_nx_graph.nodes[u][NNCFNode.NODE_NAME_ATTR]
            memory_consumption_fp_model[u_node_name] = np.prod(shape) * fp_num_bits
            memory_consumption_compressed_model[u_node_name] = np.prod(shape) * num_bits
        try:
            stats.max_fp32_activation_size = max(memory_consumption_fp_model.values()) / 2**23
            stats.max_compressed_activation_size = max(memory_consumption_compressed_model.values()) / 2**23
        except ValueError:
            stats.max_fp32_activation_size = 0
            stats.max_compressed_activation_size = 0
        return stats

    def _get_precision_for_activation_tensor(self, u_node: str, v_node: str, original_nx_graph: nx.DiGraph) -> int:
        pred_u_nodes = original_nx_graph._pred[u_node]
        precision_enter_activation_tensor = max(
            [0] + [original_nx_graph.edges[pred_u_node, u_node]["precision"] for pred_u_node in pred_u_nodes]
        )
        u_node_name = original_nx_graph.nodes[u_node][NNCFNode.NODE_NAME_ATTR]
        module = self._compressed_model.nncf.get_containing_module(u_node_name)
        if is_nncf_module(module):
            quantizer = self._get_weight_quantizer_for_module(module)
            if quantizer is not None:
                precision = max(quantizer.num_bits, precision_enter_activation_tensor)
            else:
                precision = 32
            return precision

        for aq_id, aq in self._non_weight_quantizers.items():
            if u_node_name == aq_id.target_node_name:
                precision = aq.quantizer_module_ref.num_bits
                break
        else:
            precision = precision_enter_activation_tensor
        return precision

    def _get_weight_quantizer_for_module(self, module: torch.nn.Module) -> Optional[BaseQuantizer]:
        for wq_info in self._weight_quantizers.values():
            if wq_info.quantized_module is module:
                return wq_info.quantizer_module_ref
        return None


class ShareEdgesQuantizedDataPathStatisticsCollector(StatisticsCollector):
    """
    This metric calculates the percentage of quantized edges relative to the total number of edges
    in the original network graph. "Quantized edge" is an edge representing a quantized activation tensor.
    """

    QUANTIZED_EDGES_ATTR = "quantized"
    PASSED_EDGES_ATTR = "passed"
    NODES_GRAPH_ATTR = "nodes"
    IS_MERGED_GRAPH_ATTR = "is_merged"

    def __init__(
        self, compressed_model: NNCFNetwork, qctrl: "QuantizationController", target_device: TargetDevice  # noqa: F821
    ):  # noqa: E501, F821
        self._compressed_model = compressed_model
        self._qctrl = qctrl
        self.stats = QuantizationConfigurationStatistics(0, 0)
        self._target_device = target_device

    def collect(self) -> QuantizationConfigurationStatistics:
        merged_original_graph = self.get_merged_original_graph_with_patterns(
            self._compressed_model.nncf.get_original_graph()
        )
        self.stats.quantized_edges_in_cfg = 0
        nx.set_edge_attributes(merged_original_graph, False, self.QUANTIZED_EDGES_ATTR)
        nx.set_edge_attributes(merged_original_graph, False, self.PASSED_EDGES_ATTR)

        input_nodes = [node for node in merged_original_graph.nodes if len(merged_original_graph._pred[node]) == 0]
        queue = deque()
        for input_node in input_nodes:
            next_nodes = merged_original_graph._succ[input_node]
            for next_node_key in next_nodes:
                edge = merged_original_graph.edges[input_node, next_node_key]
                edge[self.PASSED_EDGES_ATTR] = True
                edge[self.QUANTIZED_EDGES_ATTR] = True
                self.stats.quantized_edges_in_cfg += 1
                queue.appendleft(next_node_key)
        visited_nodes = {}

        while len(queue) != 0:
            node_key = queue.pop()
            if node_key in visited_nodes:
                continue
            if self._all_enter_edges_in_node_of_type(merged_original_graph, node_key, self.PASSED_EDGES_ATTR):
                visited_nodes[node_key] = True
                node = merged_original_graph.nodes[node_key]
                if node[self.IS_MERGED_GRAPH_ATTR]:
                    last_node = node[self.NODES_GRAPH_ATTR][-1]
                    node_name = str(last_node[NNCFNode.NODE_NAME_ATTR])
                    matched = False
                    for aq_info in self._qctrl.non_weight_quantizers.values():
                        for target_point in aq_info.affected_insertions:
                            if node_name == target_point.target_node_name:
                                matched = True
                                break
                    if matched:
                        self._marking_edges(merged_original_graph, node_key, queue)
                    else:
                        self._marking_edges(merged_original_graph, node_key, queue, False)
                else:
                    node_name = str(node[NNCFNode.NODE_NAME_ATTR])

                    matched = False
                    for aq_key in self._compressed_model.nncf.external_quantizers:
                        if node_name in aq_key:
                            matched = True
                            break
                    if matched:
                        self._marking_edges(merged_original_graph, node_key, queue)
                    else:
                        is_op_non_change_precision_activation_tensor = True
                        node_metatype = node[NNCFNode.METATYPE_ATTR]
                        is_op_non_change_precision_activation_tensor = (
                            node_metatype not in DEFAULT_PT_QUANT_TRAIT_TO_OP_DICT[QuantizationTrait.INPUTS_QUANTIZABLE]
                        )

                        status = is_op_non_change_precision_activation_tensor and self._all_enter_edges_in_node_of_type(
                            merged_original_graph, node_key, self.QUANTIZED_EDGES_ATTR
                        )
                        self._marking_edges(merged_original_graph, node_key, queue, status)
            else:
                queue.appendleft(node_key)
        self.num_merged_original_graph_edges = len(merged_original_graph.edges)
        self.stats.total_edges_in_cfg = self.num_merged_original_graph_edges
        return self.stats

    def _all_enter_edges_in_node_of_type(self, graph, node_key, type_edge):
        prev_nodes = graph._pred[node_key]
        retval = True
        for prev_node_key in prev_nodes:
            edge = graph.edges[prev_node_key, node_key]
            if not edge[type_edge]:
                retval = False
                break
        return retval

    def _marking_edges(self, graph, node_key, queue, mark=True):
        next_nodes = graph._succ[node_key]
        for next_node_key in next_nodes:
            edge = graph.edges[node_key, next_node_key]
            edge[self.QUANTIZED_EDGES_ATTR] = mark
            edge[self.PASSED_EDGES_ATTR] = True
            queue.appendleft(next_node_key)
            if mark:
                self.stats.quantized_edges_in_cfg += 1

    def get_merged_original_graph_with_patterns(self, original_graph: PTNNCFGraph):
        pattern = PatternsManager.get_full_hw_pattern_graph(backend=BackendType.TORCH, device=self._target_device)

        matches = find_subgraphs_matching_pattern(original_graph._nx_graph, pattern)
        merged_graph = deepcopy(original_graph._nx_graph)
        nx.set_node_attributes(merged_graph, False, self.IS_MERGED_GRAPH_ATTR)
        for match in matches:
            if len(match) == 1:
                continue

            input_node_key = match[0]
            output_node_key = match[-1]
            in_edges = list(merged_graph.in_edges(input_node_key))
            out_edges = list(merged_graph.out_edges(output_node_key))

            in_edge_copies_dict = {}
            for in_edge_key in in_edges:
                in_edge_copies_dict[in_edge_key] = deepcopy(merged_graph.edges[in_edge_key])
            out_edge_copies_dict = {}
            for out_edge_key in out_edges:
                out_edge_copies_dict[out_edge_key] = deepcopy(merged_graph.edges[out_edge_key])

            merged_node_key = ""
            merged_nodes = []
            for node_key in match:
                merged_node_key += node_key + "\n"

                merged_nodes.append(original_graph._nx_graph.nodes[node_key])
                merged_graph.remove_node(node_key)
            merged_node_attrs = {
                NNCFNode.KEY_NODE_ATTR: merged_node_key,
                self.NODES_GRAPH_ATTR: merged_nodes,
                self.IS_MERGED_GRAPH_ATTR: True,
            }
            merged_graph.add_node(merged_node_key, **merged_node_attrs)
            for in_edge_key, in_edge_attrs in in_edge_copies_dict.items():
                merged_graph.add_edge(in_edge_key[0], merged_node_key, **in_edge_attrs)
            for out_edge_key, out_edge_attrs in out_edge_copies_dict.items():
                merged_graph.add_edge(merged_node_key, out_edge_key[1], **out_edge_attrs)

        return merged_graph

    @staticmethod
    def visualize_marked_graph(merged_original_graph):
        out_graph = nx.DiGraph()
        for node_key, _ in merged_original_graph.nodes.items():
            out_graph.add_node(node_key)
        for u, v in merged_original_graph.edges:
            edge = merged_original_graph.edges[u, v]
            if edge[ShareEdgesQuantizedDataPathStatisticsCollector.QUANTIZED_EDGES_ATTR]:
                attrs = {"color": "blue"}
            out_graph.add_edge(u, v, **attrs)
        return out_graph



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import queue
from copy import copy
from copy import deepcopy
from functools import partial
from typing import Callable, Dict, List, Optional, Tuple

import numpy as np
import torch
from torch import nn
from torch import optim

from nncf.config.config import NNCFConfig
from nncf.config.schemata.defaults import PRUNING_LEGR_GENERATIONS
from nncf.config.schemata.defaults import PRUNING_LEGR_MUTATE_PERCENT
from nncf.config.schemata.defaults import PRUNING_LEGR_NUM_SAMPLES
from nncf.config.schemata.defaults import PRUNING_LEGR_POPULATION_SIZE
from nncf.config.schemata.defaults import PRUNING_LEGR_SIGMA_SCALE
from nncf.torch.utils import get_filters_num


class EvolutionOptimizer:
    """
    Class for optimizing ranking coefficients for the model with evolution algorithm (agent).
    The evolution algorithm works as follows:
    1. For the first population_size steps it generates and returns random actions (generated with some prior
    information). For every action, it gets a reward (some measure whether this action is good or not). During these
    generations all action - reward pairs saving to the population.
    2. During remaining (generations - population_size) generations it predict action by next scheme:
        - Choosing random num_samples actions from population
        - Choosing the best one from sampled and mutate it
        - Return the resulting action
    During this generation's action - reward pairs saving by updating oldest actions in population.

    After all generations, the best action (with the best reward value) is returned.
    """

    def __init__(self, initial_filter_norms: Dict, hparams: Dict, random_seed: int):
        """
        :param initial_filter_norms: Initial filter norms needed to get std and var of filter norms in each leyer.
        :param hparams: hyperparams of the Optimizer, can contain population_size, num_generations, num_samples,
        mutate_percent, sigma_scale
        :param random_seed: random seed, thet should be set during action generation for reproducibility
        """
        self.random_seed = random_seed
        # Optimizer hyper-params
        self.population_size = hparams.get("population_size", PRUNING_LEGR_POPULATION_SIZE)
        self.num_generations = hparams.get("num_generations", PRUNING_LEGR_GENERATIONS)
        self.num_samples = hparams.get("num_samples", PRUNING_LEGR_NUM_SAMPLES)
        self.mutate_percent = hparams.get("mutate_percent", PRUNING_LEGR_MUTATE_PERCENT)
        self.scale_sigma = hparams.get("sigma_scale", PRUNING_LEGR_SIGMA_SCALE)
        self.max_reward = -np.inf
        self.mean_rewards = []

        self.indexes_queue = queue.Queue(self.population_size)
        self.oldest_index = None
        self.population_rewards = np.zeros(self.population_size)
        self.population = [None for i in range(self.population_size)]

        self.best_action = None
        self.last_action = None
        self.num_layers = len(initial_filter_norms)
        self.layer_keys = np.array(list(initial_filter_norms.keys()))
        self.initial_norms_stats = {}
        for key in self.layer_keys:
            layer_norms = initial_filter_norms[key].cpu().detach().numpy()
            self.initial_norms_stats[key] = {"mean": np.mean(layer_norms), "std": np.std(layer_norms)}

        self.cur_state = None
        self.cur_reward = None
        self.cur_episode = None
        self.cur_info = None

    def get_best_action(self):
        return self.best_action

    def _save_episode_info(self, reward: float) -> None:
        """
        Saving episode information: action-reward pairs and updating best_action/reward variables if needed.
        :param reward: reward for the current episode
        """
        # Update best action and reward if needed
        if reward > self.max_reward:
            self.best_action = self.last_action
            self.max_reward = reward

        if self.cur_episode < self.population_size:
            self.indexes_queue.put(self.cur_episode)
            self.population[self.cur_episode] = self.last_action
            self.population_rewards[self.cur_episode] = reward
        else:
            self.indexes_queue.put(self.oldest_index)
            self.population[self.oldest_index] = self.last_action
            self.population_rewards[self.oldest_index] = reward

    def _predict_action(self) -> Dict:
        """
        Predict action for the current episode. Works as described above.
        :return: new generated action
        """
        np.random.seed(self.random_seed)
        episode_num = self.cur_episode
        action = {}

        if episode_num < self.population_size - 1:
            # During first population_size generations, generates random actions
            for key in self.layer_keys:
                scale = np.exp(np.random.normal(0, self.scale_sigma))
                shift = np.random.normal(0, self.initial_norms_stats[key]["std"])
                action[key] = (scale, shift)
        elif episode_num == self.population_size - 1:
            # Adding identity action to population
            for key in self.layer_keys:
                action[key] = (1, 0)
        else:
            step_size = 1 - (float(episode_num) / (self.num_generations * 1.25))  # Rename this
            self.mean_rewards.append(np.mean(self.population_rewards))

            # 1. Sampling  num_samples actions from population and choosing the best one
            sampled_idxs = np.random.choice(len(self.population_rewards), self.num_samples)
            sampled_rewards = self.population_rewards[sampled_idxs]
            best_action = self.population[sampled_idxs[np.argmax(sampled_rewards)]]

            # 2. Mutate best action
            mutate_num = int(self.mutate_percent * self.num_layers)
            mutate_idxs = np.random.choice(self.layer_keys, mutate_num)
            for key in self.layer_keys:
                scale, shift = 1, 0
                if key in mutate_idxs:
                    scale = np.exp(np.random.normal(0, self.scale_sigma * step_size))
                    shift = np.random.normal(0, self.initial_norms_stats[key]["std"])
                action[key] = (scale * best_action[key][0], shift + best_action[key][1])
            self.oldest_index = self.indexes_queue.get()
        return action

    def ask(self, episode_num: int) -> Dict:
        """
        Predict and returns action for the last told episode information: state, reward, episode_num and info
        :return: predicted action
        """
        self.cur_episode = episode_num
        action = self._predict_action()
        self.last_action = action
        return action

    def tell(self, state: torch.Tensor, reward: float, end_of_episode: bool, episode_num: int, info: List) -> None:
        """
        Getting info about episode step and save it every end of episode
        """
        # Saving state, reward and info from the current step
        self.cur_state = state
        self.cur_reward = reward
        self.cur_episode = episode_num
        self.cur_info = info

        if end_of_episode:
            self._save_episode_info(reward)


class LeGREvolutionEnv:
    """
    Environment class for optimizing the accuracy of the pruned model with different ranking coefficients.
    During 'step' environment doing step with received action calculates current reward and useful info and return it
    During 'reset' resetting Pruner and environment params changed during iteration.
    """

    def __init__(
        self,
        filter_pruner: "LeGRPruner",
        model: nn.Module,
        train_loader: torch.utils.data.DataLoader,
        val_loader: torch.utils.data.DataLoader,
        train_fn: Callable,
        train_optimizer: Optional[torch.optim.Optimizer],
        val_fn: Callable,
        config: NNCFConfig,
        train_steps: int,
        pruning_max: float,
    ):
        """
        :param filter_pruner: LeGRPruner, should have an interface for pruning model and resetting pruner.
        :param model: target model for which ranking coefficients are trained
        :param train_loader: data loader for training the model
        :param val_loader: data loader for validating the model
        :param train_fn: callable for training the model
        :param train_optimizer: optional, optimizer for training the model
        :param val_fn: callable for validation of the model, returns acc, loss
        :param config: NNCF config for model compression
        :param train_steps: number of training steps to evaluate action (ranking coefficients set)
        :param pruning_max: pruning level for the model
        """
        self.loss_as_reward = True
        self.prune_target = pruning_max
        self.steps = train_steps

        # Train/test params
        self.train_loader, self.val_loader = train_loader, val_loader
        self.train_fn = train_fn
        self.train_optimizer = train_optimizer
        if self.train_optimizer is None:
            # Default optimizer when the user did not provide a custom optimizer
            self.train_optimizer = partial(optim.SGD, lr=1e-2, momentum=0.9, weight_decay=5e-4, nesterov=True)
        self.validate_fn = val_fn
        self.config = config

        self.filter_pruner = filter_pruner
        self.model = model

    def reset(self) -> Tuple[torch.Tensor, List]:
        """
        Resetting pruner params (all changes in the model made by training) and environment params changed during
        the step.
        :return: tuple with state and info : full flops in the model and number of flops that is rest in the model
        """
        self.filter_pruner.reset()
        self.model.eval()

        self.full_flops = self.filter_pruner.get_full_flops_number_in_model()
        self.rest = self.full_flops
        self.last_act = None
        return torch.zeros(1), [self.full_flops, self.rest]

    def _train_steps(self, steps: int) -> None:
        """
        Training model with train_fn for received steps number.
        :param steps: number of model training steps
        """
        optimizer = self.train_optimizer(self.model.parameters())
        self.train_fn(self.train_loader, self.model, optimizer, self.filter_pruner, steps)

    def _get_reward(self) -> Tuple[float, float, float]:
        """
        Validating model with validate_fn and return result in format: (acc, loss)
        """
        return self.validate_fn(self.model, self.val_loader)

    def step(self, action: Dict) -> Tuple[torch.Tensor, float, bool, List]:
        """
        1. Getting action (ranking coefficients)
        2. Making step with this action - prune model with ranking coefficients
        3. Getting a reward for this action- train model for some steps and validate it
        4. Returning new state (for current settings state is default and not used), reward,
         whether the episode is over or not (for current settings an episode is over after every step) and additional
          info (full flops in model and flops left in the model)
        :param action: ranking coefficients
        """
        self.last_act = action
        new_state = torch.zeros(1)

        reduced = self.filter_pruner.prune(self.prune_target, action)
        self._train_steps(self.steps)

        acc, loss = self._get_reward()
        if self.loss_as_reward:
            reward = -loss
        else:
            reward = acc
        done = 1
        info = [self.full_flops, reduced]
        return new_state, reward, done, info


class LeGRPruner:
    """
    Wrapper for pruning controller with a simplified interface, allowing prune model with received ranking coefficients
    and resetting all changes in the model made by the environment.
    """

    def __init__(self, filter_pruner_ctrl: "FilterPruningController", model: nn.Module):  # noqa: F821
        self.filter_pruner = filter_pruner_ctrl
        self.scheduler = copy(self.filter_pruner.scheduler)
        self.model = model
        self.model_params_copy = None
        self._save_model_weights()
        self.init_filter_norms = {
            node.node_name: self.filter_pruner.filter_importance(node.module.weight)
            for node in self.filter_pruner.pruned_module_groups_info.get_all_nodes()
        }

    def loss(self) -> float:
        """
        :return: loss for pruning algorithm
        """
        return self.filter_pruner.loss()

    def _save_model_weights(self) -> None:
        """
        Saving copy of all model parameters
        """
        self.model_params_copy = deepcopy(self.model.state_dict())

    def _restore_model_weights(self):
        """
        Restoring saved original model parameters to discard any changes in model weights.
        """
        self.model.load_state_dict(self.model_params_copy)

    def _reset_masks(self) -> None:
        """
        Resetting masks for all pruned nodes
        """
        for minfo in self.filter_pruner.pruned_module_groups_info.get_all_nodes():
            new_mask = torch.ones(get_filters_num(minfo.module)).to(minfo.module.weight.device)
            self.filter_pruner.set_mask(minfo, new_mask)

    def reset(self) -> None:
        """
        Resetting all changes made in the model (and model masks during environment step) by restoring the original
        model weights, resetting masks.
        """
        self._restore_model_weights()
        self._reset_masks()
        self.scheduler = copy(self.filter_pruner.scheduler)

    def get_full_flops_number_in_model(self) -> float:
        return self.filter_pruner.full_flops

    def prune(self, flops_pruning_target: float, ranking_coeffs: Dict) -> None:
        """
        Prune target model to flops pruning target with ranking_coeffs.
        :param flops_pruning_target: pruning target for the model pruning
        :param ranking_coeffs: ranking coefficients, that will be used for layers ranking during pruning
        """
        self.filter_pruner.ranking_coeffs = ranking_coeffs
        self.filter_pruner.set_pruning_level(flops_pruning_target)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.transformations.command_creation import CommandCreator
from nncf.common.graph.transformations.commands import TargetType
from nncf.onnx.graph.transformations.commands import ONNXInitializerUpdateCommand
from nncf.onnx.graph.transformations.commands import ONNXQDQNodeRemovingCommand
from nncf.onnx.graph.transformations.commands import ONNXTargetPoint


def create_bias_correction_command(node: NNCFNode, bias_value: np.ndarray) -> ONNXInitializerUpdateCommand:
    """
     Creates bias correction command.

    :param node: The node in the NNCF graph that corresponds to operation with bias.
    :param bias_value: The new bias value that will be set.
    :return: The `ONNXInitializerUpdateCommand` command to update bias.
    """
    bias_port_id = node.metatype.bias_port_id
    target_point = ONNXTargetPoint(TargetType.LAYER, node.node_name, bias_port_id)
    return ONNXInitializerUpdateCommand(target_point, bias_value)


class ONNXCommandCreator(CommandCreator):
    """
    Implementation of the `CommandCreator` class for the ONNX backend.
    """

    @staticmethod
    def create_command_to_remove_quantizer(quantizer_node: NNCFNode) -> ONNXQDQNodeRemovingCommand:
        target_point = ONNXTargetPoint(TargetType.LAYER, quantizer_node.node_name, port_id=None)
        return ONNXQDQNodeRemovingCommand(target_point)

    @staticmethod
    def create_command_to_update_bias(
        node_with_bias: NNCFNode, bias_value: np.ndarray, nncf_graph: NNCFGraph
    ) -> ONNXInitializerUpdateCommand:
        return create_bias_correction_command(node_with_bias, bias_value)

    @staticmethod
    def create_command_to_update_weight(
        node_with_weight: NNCFNode, weight_value: np.ndarray, weight_port_id: int
    ) -> ONNXInitializerUpdateCommand:
        target_point = ONNXTargetPoint(TargetType.LAYER, node_with_weight.node_name, weight_port_id)
        return ONNXInitializerUpdateCommand(target_point, weight_value)

    @staticmethod
    def create_command_to_insert_bias(node_without_bias, bias_value):
        raise NotImplementedError



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from nncf.tensorflow.graph.utils import get_weight_by_name
from nncf.tensorflow.layers.custom_objects import NNCF_CUSTOM_OBJECTS
from nncf.tensorflow.layers.operation import InputType
from nncf.tensorflow.layers.operation import NNCFOperation
from nncf.tensorflow.sparsity.magnitude.functions import apply_mask


@NNCF_CUSTOM_OBJECTS.register()
class BinaryMask(NNCFOperation):
    def build(self, input_shape, input_type, name, layer):
        if input_type is not InputType.WEIGHTS:
            raise ValueError("Binary Mask operation could not be applied to input of the layer: {}".format(layer.name))

        mask = layer.add_weight(
            name + "_mask",
            shape=input_shape,
            initializer=tf.keras.initializers.Constant(1.0),
            trainable=False,
            aggregation=tf.VariableAggregation.MEAN,
        )

        return {"mask": mask}

    def call(self, inputs, weights, _):
        return apply_mask(inputs, weights["mask"])

    @staticmethod
    def get_binary_mask(op_weights):
        """
        Returns binary mask from weights of the operation.

        :param op_weights: Weights of the operaton.
        :return: Binary mask.
        """
        return op_weights["mask"]


@NNCF_CUSTOM_OBJECTS.register()
class BinaryMaskWithWeightsBackup(BinaryMask):
    def __init__(self, name: str, w_name_to_bkup: str = None):
        super().__init__(name)
        self.w_name_to_bkup = w_name_to_bkup
        self.bkup_var = None

    def build(self, input_shape, input_type, name, layer):
        self.bkup_var = self._create_bkup_weights(layer, self.w_name_to_bkup)
        return super().build(input_shape, input_type, name, layer)

    def call(self, inputs, weights, _):
        self.bkup_var.assign(tf.where(weights["mask"] > 0.5, inputs, self.bkup_var))
        return apply_mask(self.bkup_var, weights["mask"])

    @staticmethod
    def _create_bkup_weights(layer, w_name):
        var = get_weight_by_name(layer, w_name)
        bkup_var = layer.add_weight(
            w_name + "_bkup", shape=var.shape, trainable=False, aggregation=tf.VariableAggregation.MEAN
        )

        bkup_var.assign(var.read_value())
        return bkup_var

    def get_config(self):
        config = super().get_config()
        config["w_name_to_bkup"] = self.w_name_to_bkup
        return config



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import OrderedDict
from typing import Any, Dict, List

from nncf import NNCFConfig
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import SingleElasticityBuilder
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import create_elasticity_builder_from_config
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elasticity_controller import ElasticityController
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elasticity_dim import ElasticityDim
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.multi_elasticity_handler import MultiElasticityHandler
from nncf.torch.algo_selector import PT_COMPRESSION_ALGORITHMS
from nncf.torch.compression_method_api import PTCompressionAlgorithmBuilder
from nncf.torch.graph.transformations.layout import PTTransformationLayout
from nncf.torch.nncf_network import NNCFNetwork


class EBuilderStateNames:
    AVAILABLE_ELASTICITY_DIMS = "available_elasticity_dims"
    BUILDER_STATES = "builder_states"


@PT_COMPRESSION_ALGORITHMS.register("elasticity")
class ElasticityBuilder(PTCompressionAlgorithmBuilder):
    """
    Determines which modifications should be made to the original FP32 model in order to introduce elasticity
    to the model.
    """

    _state_names = EBuilderStateNames

    # NOTE: This is the order of activation elasticity dimensions when multiple of them are enabled.
    # Don't confuse with the order of adding elasticity dimension on training stages (progressiveness of
    # elasticity). For vanilla progressive shrinking the stages order is the following:
    #   1st stage: kernel
    #   2nd stage: kernel + depth
    #   3rd stage: kernel + depth + width
    # The execution order is orthogonal to this.
    # Though the order of kernel/width operations shouldn't lead to a different result mathematically,
    # there may be a minor floating-point error in the 6th sign. To make behavior stable, elastic kernel is always
    # applied after width.
    # Depth goes after width, because currently only depth handler knows about shapes after setting config by width
    # handler and don't skip some blocks if shapes on the block boundaries are not the same. Potentially, we could
    # support an alternative strategy, when width handler knows about skipped blocks and ,
    # but currently, it's not supported.
    ALL_DIMS_IN_EXECUTION_ORDER = [ElasticityDim.WIDTH, ElasticityDim.KERNEL, ElasticityDim.DEPTH]

    def __init__(self, nncf_config: NNCFConfig, should_init: bool = True):
        super().__init__(nncf_config, should_init)
        self._multi_elasticity_handler = None
        # TODO(nlyalyus): ignored/target scope is not supported (ticket 68052)
        self._ignored_scopes = self.config.get("ignored_scopes", None)
        self._target_scopes = self.config.get("target_scopes", None)
        self._multi_elasticity_handler_state = None

        all_elasticity_dims = {e.value for e in ElasticityDim}
        available_elasticity_dims_str = self._algo_config.get("available_elasticity_dims", all_elasticity_dims)
        self._available_elasticity_dims = list(map(ElasticityDim, available_elasticity_dims_str))
        self._elasticity_builders: Dict[ElasticityDim, SingleElasticityBuilder] = OrderedDict()
        self._builder_states = None

    def initialize(self, model: NNCFNetwork) -> None:
        """
        Initialize model parameters before training

        :param model: The model with additional modifications necessary to enable
            algorithm-specific compression during fine-tuning.
        """

    def get_available_elasticity_dims(self) -> List[ElasticityDim]:
        """
        :return: list of available elasticity dimensions
        """
        return self._available_elasticity_dims

    def _get_algo_specific_config_section(self) -> Dict:
        return self.config.get("bootstrapNAS", {}).get("training", {}).get("elasticity", {})

    def _build_controller(self, model: NNCFNetwork) -> "ElasticityController":
        """
        Simple implementation of building controller without setting builder state and loading controller's one.

        :param model: The model with additional modifications necessary to enable
            algorithm-specific compression during fine-tuning.
        :return: The instance of the `ElasticityController`.
        """
        return ElasticityController(model, self._algo_config, self._multi_elasticity_handler)

    def _get_transformation_layout(self, target_model: NNCFNetwork) -> PTTransformationLayout:
        sorted_elasticity_dims = list(
            filter(lambda x: x in self._available_elasticity_dims, self.ALL_DIMS_IN_EXECUTION_ORDER)
        )
        ignored_scopes = self._ignored_scopes
        target_scopes = self._target_scopes

        for elasticity_dim in sorted_elasticity_dims:
            elasticity_config = self._algo_config.get(elasticity_dim.value, {})
            elasticity_builder = create_elasticity_builder_from_config(
                elasticity_config, elasticity_dim, ignored_scopes, target_scopes
            )
            self._elasticity_builders[elasticity_dim] = elasticity_builder

        if self._builder_states is not None:
            for dim_str, builder_state in self._builder_states.items():
                dim = ElasticityDim(dim_str)
                if dim in self._elasticity_builders:
                    self._elasticity_builders[dim].load_state(builder_state)

        elasticity_handlers = OrderedDict()
        for dim, builder in self._elasticity_builders.items():
            handler = builder.build(target_model)
            elasticity_handlers[dim] = handler
        self._multi_elasticity_handler = MultiElasticityHandler(elasticity_handlers, target_model)

        layout = PTTransformationLayout()
        for handler in elasticity_handlers.values():
            commands = handler.get_transformation_commands()
            for command in commands:
                layout.register(command)
        return layout

    def _get_state_without_name(self) -> Dict[str, Any]:
        """
        Implementation of get_state that returns state without builder name.

        :return: Returns a dictionary with Python data structures
            (dict, list, tuple, str, int, float, True, False, None) that represents state of the object.
        """
        builder_states = {dim.value: builder.get_state() for dim, builder in self._elasticity_builders.items()}
        available_elasticity_dims_state = list(map(lambda x: x.value, self.get_available_elasticity_dims()))
        return {
            self._state_names.BUILDER_STATES: builder_states,
            self._state_names.AVAILABLE_ELASTICITY_DIMS: available_elasticity_dims_state,
        }

    def _load_state_without_name(self, state_without_name: Dict[str, Any]):
        """
        Implementation of load state that takes state without builder name.

        :param state_without_name: Output of `_get_state_without_name()` method.
        """
        self._builder_states = state_without_name[self._state_names.BUILDER_STATES]
        available_elasticity_dims_state = state_without_name[self._state_names.AVAILABLE_ELASTICITY_DIMS]

        # No conflict resolving with the related config options, parameters are overridden by compression state
        self._available_elasticity_dims = list(map(ElasticityDim, available_elasticity_dims_state))



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Tuple


def calculate_symmetric_level_ranges(num_bits: int, signed: bool, narrow_range: bool = False) -> Tuple[int, int]:
    """
    Calculates the numbers of the low and high quant and the number of
    quantization levels for the symmetric quantization scheme.

    :param num_bits: The bitwidth of the quantization.
    :param signed: The flag specifying type of the symmetric quantization scheme
        if it is True then the symmetric quantization scheme is the signed and
        the un-signed otherwise.
    :param narrow_range: The flag specifying quantization range if it is True
        then [1; 2^num_bits - 1] and [0; 2^num_bits - 1] otherwise.
    :return: A Tuple
        level_low - the low quant number
        level_high - the high quant number
    """
    levels = 2**num_bits

    if signed:
        level_high = (levels // 2) - 1
        level_low = -(levels // 2)
    else:
        level_high = levels - 1
        level_low = 0

    if narrow_range:
        if level_low < 0:
            level_low += 1
        else:
            level_high -= 1

    return level_low, level_high


def calculate_asymmetric_level_ranges(num_bits: int, narrow_range: bool = False) -> Tuple[int, int]:
    """
    Calculates the numbers of the low and high quant and the number of
    quantization levels for the asymmetric quantization scheme.

    :param num_bits: The bitwidth of the quantization
    :param narrow_range: The flag specifying quantization range if it is True
        then [1; 2^num_bits - 1] and [0; 2^num_bits - 1] otherwise
    :return: A Tuple
        level_low - the low quant number
        level_high - the high quant number
    """
    levels = 2**num_bits
    level_high = levels - 1
    level_low = 0

    if narrow_range:
        level_low = level_low + 1

    return level_low, level_high


def get_num_levels(level_low: int, level_high: int) -> int:
    return level_high - level_low + 1



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, List, NamedTuple, Tuple

from nncf.common.graph import NNCFNodeName
from nncf.common.logging import nncf_logger
from nncf.common.quantization.quantizer_setup import QuantizationPointId
from nncf.common.quantization.quantizer_setup import QuantizerSetupBase
from nncf.common.quantization.structs import QuantizerId
from nncf.torch.quantization.layers import BaseQuantizer


class AdjacentQuantizers(NamedTuple):
    """
    Combines activation and weight quantizers so that each quantizer is in the same group as the operation that it is
    affecting. Each quantizer that does not affect any node (e.g. if it only affects other quantizers as a topmost
    quantizer in a requantization scenario) will be placed in a separate group.
    :param: activation_quantizers   list of pairs of activation quantizers with their ids
    :param: weight_quantizers   list of pairs of weight quantizers with their ids
    """

    activation_quantizers: List[Tuple[QuantizerId, BaseQuantizer]]
    weight_quantizers: List[Tuple[QuantizerId, BaseQuantizer]]


class GroupsOfAdjacentQuantizers:
    """
    Contains groups of adjacent quantizers
    :param: weight_qp_id_per_activation_qp_id  gives a single activation quantizer for a given weight quantizer
    that directly quantize a weightable module (e.g. conv or linear)
    """

    def __init__(self):
        self.weight_qp_id_per_activation_qp_id: Dict[QuantizationPointId, QuantizationPointId] = {}
        self._quantizer_per_group_id = {}
        self._groups_of_adjacent_quantizers: List[AdjacentQuantizers] = []

    def get_group_id_for_quantizer(self, quantizer_id: QuantizerId):
        return self._quantizer_per_group_id.get(quantizer_id, None)

    def get_adjacent_quantizers_by_group_id(self, group_id):
        return (
            self._groups_of_adjacent_quantizers[group_id].weight_quantizers
            + self._groups_of_adjacent_quantizers[group_id].activation_quantizers
        )

    def __iter__(self):
        return iter(self._groups_of_adjacent_quantizers)

    def __bool__(self):
        return bool(self._groups_of_adjacent_quantizers) and bool(self._quantizer_per_group_id)

    def __getitem__(self, group_id):
        return self._groups_of_adjacent_quantizers[group_id]

    def parse_from_quantizer_setup(
        self,
        all_quantizations: Dict[QuantizerId, BaseQuantizer],
        quantizer_setup: QuantizerSetupBase,
        quantization_point_id_vs_quantizer_id: Dict[QuantizationPointId, QuantizerId],
    ):
        for group_idx, group in quantizer_setup.shared_input_operation_set_groups.items():
            act_quant_tuples: List[Tuple[QuantizerId, BaseQuantizer]] = []
            wt_quant_tuples: List[Tuple[QuantizerId, BaseQuantizer]] = []

            quantized_node_per_activation_qp_id: Dict[NNCFNodeName, QuantizationPointId] = {}
            module_scope_per_weight_qp_id: Dict[NNCFNodeName, QuantizationPointId] = {}

            for qp_id in group:
                qp = quantizer_setup.quantization_points[qp_id]
                quant_id = quantization_point_id_vs_quantizer_id[qp_id]
                quantizer_module = all_quantizations[quant_id]
                resulting_tuple = (quant_id, quantizer_module)
                if qp.is_weight_quantization_point():
                    wt_quant_tuples.append(resulting_tuple)
                    weight_quantized_module_node_name = qp.target_point.target_node_name
                    module_scope_per_weight_qp_id[weight_quantized_module_node_name] = qp_id
                elif qp.is_activation_quantization_point():
                    act_quant_tuples.append(resulting_tuple)
                    quantized_node_names = qp.directly_quantized_operator_node_names
                    quantized_node_per_activation_qp_id.update({node_name: qp_id for node_name in quantized_node_names})
                self._quantizer_per_group_id[quant_id] = group_idx

            for weight_quantized_module_node_name, w_qp_id in module_scope_per_weight_qp_id.items():
                if weight_quantized_module_node_name not in quantized_node_per_activation_qp_id:
                    nncf_logger.debug(
                        f"Module {weight_quantized_module_node_name} has quantized weights and no quantized inputs!"
                    )
                    continue
                a_qp_id = quantized_node_per_activation_qp_id[weight_quantized_module_node_name]
                if w_qp_id in self.weight_qp_id_per_activation_qp_id:
                    nncf_logger.debug(
                        f"Multiple weight quantizers per activation quantizer "
                        f"for {weight_quantized_module_node_name}"
                    )
                    continue
                self.weight_qp_id_per_activation_qp_id[w_qp_id] = a_qp_id

            adj_quants = AdjacentQuantizers(act_quant_tuples, wt_quant_tuples)
            self._groups_of_adjacent_quantizers.append(adj_quants)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# isort: off

"""
Base subpackage for NNCF PyTorch functionality.
"""

import os
from nncf import nncf_logger
from nncf.common.logging.logger import warn_bkc_version_mismatch

from nncf.version import BKC_TORCH_SPEC

import torch
from packaging import version
from packaging.specifiers import SpecifierSet

try:
    _torch_version = version.parse(version.parse(torch.__version__).base_version)
except:  # noqa: E722
    nncf_logger.debug("Could not parse torch version")
    _torch_version = version.parse("0.0.0")

if _torch_version not in SpecifierSet(BKC_TORCH_SPEC):
    warn_bkc_version_mismatch("torch", BKC_TORCH_SPEC, torch.__version__)


# Required for correct COMPRESSION_ALGORITHMS registry functioning
from nncf.torch.quantization import algo as quantization_algo
from nncf.torch.sparsity.const import algo as const_sparsity_algo
from nncf.torch.sparsity.magnitude import algo as magnitude_sparsity_algo
from nncf.torch.sparsity.rb import algo as rb_sparsity_algo
from nncf.experimental.torch.sparsity.movement import algo as movement_sparsity_algo
from nncf.torch.pruning.filter_pruning import algo as filter_pruning_algo
from nncf.torch.knowledge_distillation import algo as knowledge_distillation_algo

# Functions most commonly used in integrating NNCF into training pipelines are
# listed below for importing convenience

from nncf.torch.model_creation import create_compressed_model
from nncf.torch.model_creation import is_wrapped_model
from nncf.torch.model_creation import wrap_model
from nncf.torch.model_creation import load_from_config
from nncf.torch.checkpoint_loading import load_state
from nncf.torch.initialization import register_default_init_args
from nncf.torch.layers import register_module
from nncf.torch.dynamic_graph.patch_pytorch import register_operator
from nncf.torch.dynamic_graph.io_handling import nncf_model_input
from nncf.torch.dynamic_graph.io_handling import nncf_model_output
from nncf.torch.dynamic_graph.context import disable_tracing
from nncf.torch.dynamic_graph.context import no_nncf_trace
from nncf.torch.dynamic_graph.context import forward_nncf_trace
from nncf.torch.strip import strip
from nncf.torch.dynamic_graph.patch_pytorch import disable_patching

# NNCF relies on tracing PyTorch operations. Each code that uses NNCF
# should be executed with PyTorch operators wrapped via a call to "patch_torch_operators",
# so this call is moved to package __init__ to ensure this.
from nncf.torch.dynamic_graph.patch_pytorch import patch_torch_operators

from nncf.torch.extensions import force_build_cpu_extensions, force_build_cuda_extensions

# This is required since torchvision changes a dictionary inside of pytorch mapping
# different ops and their role in torch fx graph. Once the nncf mapping is done, it is
# represented as a different custom operation which is how it is changed in
# the said mapping. The polyfills loader is the specific file to be imported
# before making wrapping changes
if torch.__version__ >= "2.5.0":
    from torch._dynamo.polyfills import loader

if os.getenv("NNCF_EXPERIMENTAL_TORCH_TRACING") is None:
    patch_torch_operators()



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from keras import backend as backend
from keras import layers as layers
from keras.layers import Rescaling as Rescaling
from packaging import version as version
from tensorflow.python.eager import context as eager_context  # noqa: F401

from nncf.tensorflow import tensorflow_version

if tensorflow_version < version.parse("2.13"):
    from keras import engine as keras_engine  # noqa: F401
    from keras.applications import imagenet_utils as imagenet_utils
    from keras.engine.keras_tensor import KerasTensor as KerasTensor
    from keras.utils.control_flow_util import smart_cond as smart_cond
else:
    from keras.src import engine as keras_engine  # noqa: F401
    from keras.src.applications import imagenet_utils as imagenet_utils  # noqa: E501
    from keras.src.engine.keras_tensor import KerasTensor as KerasTensor  # noqa: E501
    from keras.src.utils.control_flow_util import smart_cond as smart_cond  # noqa: E501



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import tensorflow as tf

from nncf.tensorflow import tf_internals
from nncf.tensorflow.functions import logit
from nncf.tensorflow.layers.custom_objects import NNCF_CUSTOM_OBJECTS
from nncf.tensorflow.layers.operation import InputType
from nncf.tensorflow.layers.operation import NNCFOperation
from nncf.tensorflow.layers.wrapper import NNCFWrapper
from nncf.tensorflow.sparsity.magnitude.functions import apply_mask
from nncf.tensorflow.sparsity.rb.functions import binary_mask
from nncf.tensorflow.sparsity.rb.functions import calc_rb_binary_mask
from nncf.tensorflow.sparsity.rb.functions import st_binary_mask


@NNCF_CUSTOM_OBJECTS.register()
class RBSparsifyingWeight(NNCFOperation):
    def __init__(self, name: str, eps: float = 1e-6):
        """
        :param name: Model scope unique operation name.
        :param eps: Minimum value and the gap from the maximum value
            in the mask.
        """
        super().__init__(name)
        self.eps = eps

    def build(self, input_shape, input_type: InputType, name: str, layer: NNCFWrapper):
        """
        :param input_shape: Shape of weights which needs to be sparsifyed.
        :param input_type: Type of operation input, must be InputType.WEIGHTS.
        :param name: Name of weight attribute which needs to be sparsifyed.
        :param layer: Layer which needs to be sparsifyed.
        """
        if input_type is not InputType.WEIGHTS:
            raise ValueError(
                "RB Sparsity mask operation could not be applied to input of the layer: {}".format(layer.name)
            )

        mask = layer.add_weight(
            name + "_mask",
            shape=input_shape,
            initializer=tf.keras.initializers.Constant(logit(0.99)),
            trainable=True,
            aggregation=tf.VariableAggregation.MEAN,
        )

        trainable = layer.add_weight(
            name + "_trainable", initializer=tf.keras.initializers.Constant(True), trainable=False, dtype=tf.bool
        )

        seed = layer.add_weight(
            name + "_seed",
            shape=(2,),
            initializer=tf.keras.initializers.Constant(np.random.randint(size=(2,), low=-(2**31), high=2**31 - 1)),
            trainable=False,
            dtype=tf.int32,
        )

        return {
            "mask": mask,
            "trainable": trainable,
            "seed": seed,
        }

    def call(self, inputs, weights, training: tf.constant):
        """
        Apply rb sparsity mask to given weights.

        :param inputs: Target weights to sparsify.
        :param weights: Operation weights contains
            `mask` and param `trainable`.
        :param training: True if operation called in training mode
            else False
        """
        true_fn = lambda: apply_mask(inputs, self._calc_rb_binary_mask(weights))
        false_fn = lambda: apply_mask(inputs, binary_mask(weights["mask"]))
        return tf_internals.smart_cond(
            training,
            true_fn=lambda: tf_internals.smart_cond(weights["trainable"], true_fn=true_fn, false_fn=false_fn),
            false_fn=false_fn,
        )

    def _calc_rb_binary_mask(self, op_weights):
        new_seed = tf.random.stateless_uniform((2,), seed=op_weights["seed"], minval=-(2**31), maxval=2**31 - 1)
        new_seed = tf.cast(new_seed, tf.int32)
        op_weights["seed"].assign(new_seed)
        return calc_rb_binary_mask(op_weights["mask"], op_weights["seed"], self.eps)

    def freeze(self, op_weights):
        """
        Freeze rb mask from operation weights.

        :param op_weights: Operation weights.
        """
        op_weights["trainable"].assign(False)

    @staticmethod
    def loss(op_weights):
        """
        Return count of non zero weight in mask.

        :param op_weights: Operation weights.
        """
        return tf.reduce_sum(st_binary_mask(op_weights["mask"]))

    @staticmethod
    def get_mask(op_weights):
        """
        Return mask weight from operation weights.

        :param op_weights: Operation weights.
        """
        return op_weights["mask"]

    @staticmethod
    def get_binary_mask(op_weights):
        """
        Returns binary mask from weights of the operation.

        :param op_weights: Weights of the operaton.
        :return: Binary mask.
        """
        return binary_mask(op_weights["mask"])

    @staticmethod
    def get_trainable_weight(op_weights):
        """
        Return trainable weight from operation weights.

        :param op_weights: Operation weights.
        """
        return op_weights["trainable"]

    def get_config(self):
        config = super().get_config()
        config["eps"] = self.eps
        return config



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from functools import reduce
from typing import Any, Callable, Optional

import torch
from torch import nn

from nncf.common.logging import nncf_logger
from nncf.torch.compression_method_api import PTCompressionLoss
from nncf.torch.nested_objects_traversal import NestedObjectIndex
from nncf.torch.nncf_network import NNCFNetwork


class KnowledgeDistillationLoss(PTCompressionLoss):
    """
    Doesn't not directly compute knowledge distillation loss values but has access to them through
    KnowledgeDistillationLossHandler. Notice that knowledge distillation loss is computed between results of original
    model and compressed model inferences with latest inputs. Provides KnowledgeDistillationLossHandler with kd original
    model (to distill from), storage device and function to calculate knowledge distillation loss.
    """

    def __init__(
        self, target_model: NNCFNetwork, original_model: nn.Module, kd_type: str, scale: float, temperature: float
    ):
        super().__init__()
        original_model.train()
        if kd_type == "softmax":

            def kd_loss_fn(teacher_output: torch.Tensor, student_output: torch.Tensor):
                if len(student_output.shape) != 2 or len(teacher_output.shape) != 2:
                    nncf_logger.debug(
                        f"Incompatible number of dimensions of the model output tensor for softmax KD "
                        f"(student - {student_output.shape}, "
                        f"teacher - {teacher_output.shape}, "
                        f"number of dims for both should be == 2) - ignoring!"
                    )
                    return torch.zeros([1]).to(student_output.device)
                return (
                    scale
                    * -(
                        nn.functional.log_softmax(student_output / temperature, dim=1)
                        * nn.functional.softmax(teacher_output / temperature, dim=1)
                    ).mean()
                    * (student_output.shape[1] * temperature * temperature)
                )

        else:

            def kd_loss_fn(teacher_output: torch.Tensor, student_output: torch.Tensor):
                mse = torch.nn.MSELoss()
                if len(teacher_output.shape) < 2:
                    nncf_logger.debug(
                        f"Incompatible number of dimensions of the model output tensor for MSE KD "
                        f"(student - {student_output.shape}, "
                        f"teacher - {teacher_output.shape}, "
                        f"number of dims {len(student_output.shape)} should be > 1) (most likely loss) - ignoring!"
                    )
                    return torch.zeros([1]).to(student_output.device)
                return scale * mse(teacher_output, student_output)

        self._kd_loss_handler = target_model.nncf.create_knowledge_distillation_loss_handler(
            original_model, partial(KnowledgeDistillationLoss._calculate, kd_loss_fn=kd_loss_fn)
        )

    @staticmethod
    def _calculate(
        compressed_model_outputs: Any,
        orig_model_outputs: Any,
        kd_loss_fn: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    ) -> Optional[torch.Tensor]:
        """
        Calculates knowledge distillation loss value from compressed_model_outputs and orig_model_outputs. First uses
        nested_object_paths_generator to unpack input containers and numerate contents inside them.
        Than checks compressed_model_outputs unpacked container for loss tensors (requires_grad=True)
        and maps extracted structure of loss tensors to orig_model_outputs.
        Finally computes knowledge distillation loss with extracted loss tensors.

        :param compressed_model_outputs: Output tensors of compressed model can be any type of container with
            deterministic traversal.
        :param orig_model_outputs: Output tensors of original model (used for distillation) can be any type of
            container with deterministic traversal.
        :return: knowledge distillation loss value
        """

        compressed_model_outputs_nested_obj_indexing = NestedObjectIndex([compressed_model_outputs])
        orig_model_outputs_nested_obj_indexing = NestedObjectIndex([orig_model_outputs])
        compressed_model_loss_outputs_nested_obj_indexing = list(
            filter(
                lambda x: KnowledgeDistillationLoss._is_loss(x.getter()),
                compressed_model_outputs_nested_obj_indexing.get_flat_nested_obj_indexing(),
            )
        )
        compressed_model_loss_outputs = list(
            map(lambda x: x.getter(), compressed_model_loss_outputs_nested_obj_indexing)
        )

        def match_fn(obj):
            for x in compressed_model_loss_outputs_nested_obj_indexing:
                if x.path == obj.path:
                    return True
            return False

        orig_model_loss_outputs = list(
            map(
                lambda x: x.getter(),
                filter(match_fn, orig_model_outputs_nested_obj_indexing.get_flat_nested_obj_indexing()),
            )
        )

        if len(orig_model_loss_outputs) != len(compressed_model_loss_outputs):
            nncf_logger.warning(
                f"KD: mismatch in the number of detected loss tensors in return value between original "
                f"and compressed models;\n"
                f"original has {len(orig_model_loss_outputs)} loss tensors,\n"
                f"compressed has {len(compressed_model_loss_outputs)}"
            )
        if not orig_model_loss_outputs:
            nncf_logger.warning("KD: no loss outputs detected in original model, knowledge distillation not possible")
            return None
        if not compressed_model_loss_outputs:
            nncf_logger.warning("KD: no loss outputs detected in compressed model, knowledge distillation not possible")
            return None
        return reduce(
            lambda kd_loss, loss_tensors: kd_loss + kd_loss_fn(loss_tensors[0], loss_tensors[1]),
            zip(orig_model_loss_outputs, compressed_model_loss_outputs),
            torch.zeros([], device=orig_model_loss_outputs[0].device),
        )

    @staticmethod
    def _is_loss(obj):
        if not isinstance(obj, torch.Tensor):
            return False
        if obj.requires_grad:
            return True
        return False

    def forward(self) -> torch.Tensor:
        """
        Gets knowledge distillation loss values from KnowledgeDistillationLossHandler, averages them in case of
        DataParallel execution (loss values for mini-batches) and frees up KnowledgeDistillationLossHandler loss values
        storage space.

        :return: Differentiable knowledge distillation loss value
        """
        loss = self._kd_loss_handler.get_kd_loss()
        for idx, _ in enumerate(loss):
            loss[idx] = loss[idx].unsqueeze(0)
        output = torch.cat(loss).mean()
        return output

    def statistics(self, quickly_collected_only=False):
        return {}



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

ONLINE_DOCS_ROOT = "https://github.com/openvinotoolkit/nncf/tree/develop/"
SCHEMA_VISUALIZATION_URL = "https://openvinotoolkit.github.io/nncf/"

ADAPTIVE_COMPRESSION_LEVEL_TRAINING_MODE_NAME_IN_CONFIG = "adaptive_compression_level"
EARLY_EXIT_TRAINING_MODE_NAME_IN_CONFIG = "early_exit"
EXPERIMENTAL_QUANTIZATION_ALGO_NAME_IN_CONFIG = "experimental_quantization"
BOOTSTRAP_NAS_ALGO_NAME_IN_CONFIG = "bootstrapNAS"
CONST_SPARSITY_ALGO_NAME_IN_CONFIG = "const_sparsity"
FILTER_PRUNING_ALGO_NAME_IN_CONFIG = "filter_pruning"
KNOWLEDGE_DISTILLATION_ALGO_NAME_IN_CONFIG = "knowledge_distillation"
MAGNITUDE_SPARSITY_ALGO_NAME_IN_CONFIG = "magnitude_sparsity"
MOVEMENT_SPARSITY_ALGO_NAME_IN_CONFIG = "movement_sparsity"
QUANTIZATION_ALGO_NAME_IN_CONFIG = "quantization"
RB_SPARSITY_ALGO_NAME_IN_CONFIG = "rb_sparsity"

ALGO_NAME_VS_README_URL = {
    QUANTIZATION_ALGO_NAME_IN_CONFIG: "docs/compression_algorithms/Quantization.md",
    FILTER_PRUNING_ALGO_NAME_IN_CONFIG: "docs/compression_algorithms/Pruning.md",
    MAGNITUDE_SPARSITY_ALGO_NAME_IN_CONFIG: "docs/compression_algorithms/Sparsity.md",
    RB_SPARSITY_ALGO_NAME_IN_CONFIG: "docs/compression_algorithms/Sparsity.md",
    CONST_SPARSITY_ALGO_NAME_IN_CONFIG: "docs/compression_algorithms/Sparsity.md",
    KNOWLEDGE_DISTILLATION_ALGO_NAME_IN_CONFIG: "docs/compression_algorithms/KnowledgeDistillation.md",
}



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, List, Tuple

import torch.fx

import nncf.torch.graph.operator_metatypes as om
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.quantization.quantizer_propagation.structs import QuantizationTrait
from nncf.common.tensor_statistics.statistic_point import StatisticPoint
from nncf.experimental.common.tensor_statistics.collectors import AbsMaxReducer
from nncf.experimental.common.tensor_statistics.collectors import MaxAggregator
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.experimental.torch.fx.commands import FXApplyTransformationCommand
from nncf.experimental.torch.fx.node_utils import get_tensor_constant_from_node
from nncf.experimental.torch.fx.transformations import constant_update_transformation_builder
from nncf.experimental.torch.fx.transformations import get_graph_node_by_name
from nncf.experimental.torch.fx.transformations import module_insertion_transformation_builder
from nncf.quantization.algorithms.smooth_quant.backend import SmoothQuantAlgoBackend
from nncf.tensor import Tensor
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.model_graph_manager import get_const_node
from nncf.torch.quantization.default_quantization import DEFAULT_PT_QUANT_TRAIT_TO_OP_DICT

PT_PRE_LAYER_TARGET_TYPE = TargetType.OPERATOR_PRE_HOOK


class FXSQMultiply(torch.nn.Module):
    def __init__(self, scale: torch.Tensor):
        super().__init__()
        self._scale_value = scale

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.mul(x, self._scale_value)


class FXSmoothQuantAlgoBackend(SmoothQuantAlgoBackend):
    @property
    def convolution_metatypes(self) -> List[OperatorMetatype]:
        return [
            om.PTConv1dMetatype,
            om.PTConv2dMetatype,
            om.PTConv3dMetatype,
        ]

    @property
    def matmul_metatypes(self) -> List[OperatorMetatype]:
        return [om.PTLinearMetatype]

    @property
    def quantize_agnostic_metatypes(self) -> List[OperatorMetatype]:
        return DEFAULT_PT_QUANT_TRAIT_TO_OP_DICT[QuantizationTrait.QUANTIZATION_AGNOSTIC]

    @staticmethod
    def pre_layer_target_type() -> TargetType:
        return PT_PRE_LAYER_TARGET_TYPE

    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> PTTargetPoint:
        return PTTargetPoint(target_type, target_node_name, input_port_id=port_id)

    @staticmethod
    def is_node_with_weights(node: NNCFNode) -> bool:
        # Metatypes of linears and convolutions guarantee
        # all nodes with the metatypes have weights, we can skip
        # this check by returning True.
        return True

    @staticmethod
    def get_activations_port_id(node: NNCFNode, nncf_graph: NNCFGraph) -> int:
        return 0

    @staticmethod
    def get_abs_max_channel_collector(
        num_samples: int, stats_reduction_axes: Tuple[int], inplace: bool, branch_key: str
    ) -> TensorCollector:
        collector = TensorCollector()
        reducer = AbsMaxReducer(reduction_axes=stats_reduction_axes)
        aggregator = MaxAggregator(num_samples=num_samples)
        collector.register_statistic_branch(branch_key, reducer, aggregator)
        return collector

    @staticmethod
    def get_weight_value(node_with_weight: NNCFNode, model: torch.fx.GraphModule, nncf_graph: NNCFGraph) -> Tensor:
        weight_node = get_const_node(node_with_weight, node_with_weight.metatype.weight_port_ids[0], nncf_graph)
        if weight_node is None:
            raise RuntimeError(f"{node_with_weight} node has no weight node.")
        graph_node = get_graph_node_by_name(model.graph, weight_node.node_name)
        weight_data = get_tensor_constant_from_node(graph_node, model)
        return Tensor(weight_data.data)

    @staticmethod
    def weight_update_command(node_with_weight: NNCFNode, weight_value: torch.Tensor) -> FXApplyTransformationCommand:
        # TODO(dlyakhov): Use input port id depending on the node metatype/attributes.
        return FXApplyTransformationCommand(
            constant_update_transformation_builder(node_with_weight, weight_value.data, input_port_id=1)
        )

    @staticmethod
    def scale_insertion_command(
        source_node: NNCFNode,
        scale_value: torch.Tensor,
        source_output_port_id: int,
        nodes: List[NNCFNode],
        scale_node_name: str,
    ) -> FXApplyTransformationCommand:
        input_port_id = 0
        target_points = []
        for node in nodes:
            target_points.append(PTTargetPoint(PT_PRE_LAYER_TARGET_TYPE, node.node_name, input_port_id=input_port_id))

        sq_multiply = FXSQMultiply(scale_value)
        return FXApplyTransformationCommand(
            module_insertion_transformation_builder(sq_multiply, target_points, scale_node_name)
        )

    @staticmethod
    def get_activation_channel_axis(node: NNCFNode, port_id: int) -> int:
        if node.metatype == om.PTLinearMetatype:
            return -1
        # TODO: Add activation axis calculation when MatMul will be supported
        return 1

    @staticmethod
    def get_weight_channel_axis(node: NNCFNode) -> int:
        # TODO: Add activation axis calculation when MatMul will be supported
        return 1

    @staticmethod
    def is_node_with_shared_weight(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        # TODO(dlyakhov): Support shared layers in TorchFX.
        # Ref: 149316
        return False

    @staticmethod
    def get_filter_fn_for_statistics(activation_port_id: int, algorithm_key: str) -> Callable[[StatisticPoint], bool]:
        def filter_func(point: StatisticPoint) -> bool:
            return (
                algorithm_key in point.algorithm_to_tensor_collectors
                and point.target_point.type == PT_PRE_LAYER_TARGET_TYPE
                and point.target_point.input_port_id == activation_port_id
            )

        return filter_func



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Any, Dict, Union

from nncf.common.graph.transformations.commands import TransformationPriority
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.torch.graph.transformations.commands import ExtraCompressionModuleType
from nncf.torch.graph.transformations.commands import PTInsertionCommand
from nncf.torch.graph.transformations.commands import PTSharedFnInsertionCommand
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.graph.transformations.commands import PTTransformationCommand
from nncf.torch.layer_utils import COMPRESSION_MODULES

COMPRESSION_STATE_ATTR = "compression_state"
SUPPORTED_COMMANDS = (PTSharedFnInsertionCommand, PTInsertionCommand)


def serialize_transformations(transformations_layout: TransformationLayout) -> Dict[str, Any]:
    """
    Serializes given transformation layout to a dict.

    :param transformation_layout: Given transformation layout.
    :return: Serialized representation of given transformation layout as a dict.
    """
    transformation_commands = []
    for command in transformations_layout.transformations:
        transformation_commands.append(serialize_command(command))

    return {COMPRESSION_STATE_ATTR: transformation_commands}


def deserialize_transformations(serialized_transformation_layout: Dict[str, Any]) -> TransformationLayout:
    """
    Deserializes given serialized transformation layout.

    :param serialized_transformation_layout: Given serialized transformation layout.
    :return: The deserialized transformation layout.
    """
    transformation_layout = TransformationLayout()
    for serialized_command in serialized_transformation_layout[COMPRESSION_STATE_ATTR]:
        command = deserialize_command(serialized_command)
        transformation_layout.register(command)

    return transformation_layout


def serialize_command(command: PTTransformationCommand) -> Dict[str, Any]:
    """
    Serializes given command layout to a dict.

    :param command: Given command.
    :return: Serialized representation of given command as a dict.
    """
    if not isinstance(command, SUPPORTED_COMMANDS):
        raise RuntimeError(f"Command type {command.__class__} is not supported.")

    serialized_transformation = dict()
    serialized_transformation["type"] = command.__class__.__name__
    if isinstance(command, PTSharedFnInsertionCommand):
        serialized_transformation["target_points"] = [point.get_state() for point in command.target_points]
        serialized_transformation["op_name"] = command.op_name
        serialized_transformation["compression_module_type"] = command.compression_module_type.value
    elif isinstance(command, PTInsertionCommand):
        serialized_transformation["target_point"] = command.target_point.get_state()

    # Check compression module is registered
    compression_module_name = command.fn.__class__.__name__
    if compression_module_name not in COMPRESSION_MODULES.registry_dict:
        raise RuntimeError(
            f"Could not serialize compression module with name {compression_module_name}."
            " Please register your module in the COMPRESSION_MODULES registry."
        )
    serialized_transformation["compression_module_name"] = compression_module_name
    serialized_transformation["fn_config"] = command.fn.get_config()
    serialized_transformation["hooks_group_name"] = command.hooks_group_name
    priority = command.priority
    serialized_transformation["priority"] = priority.value if isinstance(priority, Enum) else priority
    return serialized_transformation


def deserialize_command(serialized_command: Dict[str, Any]) -> Union[PTInsertionCommand, PTSharedFnInsertionCommand]:
    """
    Deserializes given serialized command.

    :param serialized_command: Given serialized command.
    :return: The deserialized command.
    """
    if serialized_command["type"] not in (command_cls.__name__ for command_cls in SUPPORTED_COMMANDS):
        raise RuntimeError(f"Command type {serialized_command['type']} is not supported.")

    module_cls = COMPRESSION_MODULES.get(serialized_command["compression_module_name"])
    fn = module_cls.from_config(serialized_command["fn_config"])
    priority = serialized_command["priority"]
    if priority in iter(TransformationPriority):
        priority = TransformationPriority(priority)

    if serialized_command["type"] == PTInsertionCommand.__name__:
        target_point = PTTargetPoint.from_state(serialized_command["target_point"])
        return PTInsertionCommand(
            point=target_point, fn=fn, priority=priority, hooks_group_name=serialized_command["hooks_group_name"]
        )

    target_points = [PTTargetPoint.from_state(state) for state in serialized_command["target_points"]]
    return PTSharedFnInsertionCommand(
        target_points=target_points,
        fn=fn,
        op_unique_name=serialized_command["op_name"],
        compression_module_type=ExtraCompressionModuleType(serialized_command["compression_module_type"]),
        priority=priority,
        hooks_group_name=serialized_command["hooks_group_name"],
    )



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import List, Set

import torch
import torch.fx

from nncf.common.graph.model_transformer import ModelTransformer
from nncf.experimental.torch.fx.commands import FXApplyTransformationCommand
from nncf.experimental.torch.fx.node_utils import get_graph_node_by_name
from nncf.torch.graph.transformations.commands import PTModelExtractionCommand
from nncf.torch.graph.transformations.layout import PTTransformationLayout


class FXModelTransformer(ModelTransformer):
    """
    Applies transformations upon Torch FX model.
    FXApplyTransformationCommands are made inplace,
    PTModelExtractionCommands do not change the input model.
    """

    def __init__(self, model: torch.fx.GraphModule):
        super().__init__(model)

        self._command_transformation_ordered_pairs = [
            (FXApplyTransformationCommand, self._apply_transformation),
            (PTModelExtractionCommand, self._apply_model_extraction),
        ]

    def transform(self, transformation_layout: PTTransformationLayout) -> torch.fx.GraphModule:
        """
        Transforms the target model according to given transformation layout.

        :param transformation_layout: Given transformation layout.
        :return: Target model transformered according to the given transformation layout.
        """
        # TODO(dlyakhov): Manage priorities of transformations.
        transformations = transformation_layout.transformations
        aggregated_transformations = defaultdict(list)
        for transformation in transformations:
            aggregated_transformations[transformation.__class__].append(transformation)

        model = self._model
        for transformation_cls, transformation_fn in self._command_transformation_ordered_pairs:
            transformations = aggregated_transformations[transformation_cls]
            if transformations:
                model = transformation_fn(model, transformations)

        # Do not use model.graph.eliminate_dead_code()
        # because the computational statistics code
        # is interpolated as dead code.
        model.recompile()
        return model

    @staticmethod
    def _traverse_graph(
        input_nodes: List[torch.fx.Node],
        stop_nodes: Set[torch.fx.Node],
        visited: Set[torch.fx.Node],
    ) -> None:
        """
        Traverses through the graph starting with the input nodes and
        stopping for the stop nodes and the visited nodes. As the result,
        it modifies the visited container with all nodes visited during the traverse.

        :param input_nodes: Given input nodes.
        :param stop_nodes: Given stop nodes.
        :param visited: Set of already visited nodes.
        """

        while input_nodes:
            in_node = input_nodes.pop()
            if in_node.name in visited or in_node.name in stop_nodes:
                continue

            visited.add(in_node.name)
            # Any constant is a stop op during the traversing procedure.
            if in_node.op == "get_attr":
                continue
            input_nodes.extend(in_node.all_input_nodes)
            input_nodes.extend(list(in_node.users))

    @staticmethod
    def _apply_model_extraction(
        model: torch.fx.GraphModule,
        transformations: List[PTModelExtractionCommand],
    ) -> torch.fx.GraphModule:
        """
        Returns a submodel extracted from the given model by the given transformation.

        :param model: Given model.
        :param transformations: List of one transformation which specifies
            how to retrieve a submodule from the model. In case list contains
            more than one element this function raises an assert.
        :return: Returns a submodel extracted from the given model by the given transformation.
        """

        transformation = transformations[-1]
        stop_nodes = set(transformation.input_node_names + transformation.output_node_names)
        visited = set()

        for node_name in transformation.input_node_names:
            node = get_graph_node_by_name(model.graph, node_name)
            visited.add(node.name)
            target_inputs = node.all_input_nodes[1:]
            if node.name not in transformation.output_node_names:
                target_inputs += list(node.users)
            FXModelTransformer._traverse_graph(target_inputs, stop_nodes, visited)

        for node_name in transformation.output_node_names:
            node = get_graph_node_by_name(model.graph, node_name)
            visited.add(node.name)
            if node.name not in transformation.input_node_names:
                FXModelTransformer._traverse_graph(node.all_input_nodes, stop_nodes, visited)

        extracted_graph = torch.fx.Graph()
        value_remap = {}

        def remap_fn(node: torch.fx.Node):
            return value_remap.get(node)  # noqa F821

        visited_outputs_names = []
        for node in model.graph.nodes:
            if node.name not in visited:
                continue
            if node.op == "output":
                visited_outputs_names.append(node.name)
                continue
            value_remap[node] = extracted_graph.node_copy(node, remap_fn)

        for input_name in transformation.input_node_names:
            node_with_input = get_graph_node_by_name(extracted_graph, input_name)
            with extracted_graph.inserting_before(node_with_input):
                graph_input_name = input_name + "_input"
                graph_input = extracted_graph.create_node(
                    op="placeholder",
                    target=graph_input_name,
                    name=graph_input_name,
                )

            args = list(node_with_input.args)
            args[0] = graph_input
            node_with_input.args = tuple(args)

        # Merge new output with the original output in case
        # the original output is requested in the extracted graph.
        nodes_with_output = []
        for name in transformation.output_node_names:
            nodes_with_output.append(
                name if name in visited_outputs_names else get_graph_node_by_name(extracted_graph, name)
            )

        for idx, node in enumerate(nodes_with_output):
            if isinstance(node, torch.fx.Node):
                continue
            # Current node is the original graph output.
            # Should be replaced by its arguments.
            output_node = get_graph_node_by_name(model.graph, node)
            args = output_node.args[0]
            if isinstance(args, torch.fx.Node):
                # Case of non tuple output.
                args = value_remap[args]
            else:
                # Case of tuple output.
                args = [value_remap[n] for n in args]
                # Unpack target output args in case
                # the only one arg is presented.
                if len(args) == 1:
                    args = args[0]
            nodes_with_output[idx] = args

        last_node = list(extracted_graph.nodes)[-1]
        with extracted_graph.inserting_after(last_node):
            graph_output_name = "output"
            extracted_graph.create_node(
                "output",
                graph_output_name,
                (tuple(nodes_with_output),),
                name=graph_output_name,
            )

        return torch.fx.GraphModule(model, extracted_graph)

    @staticmethod
    def _apply_transformation(
        model: torch.fx.GraphModule,
        transformations: List[FXApplyTransformationCommand],
    ) -> torch.fx.GraphModule:
        """
        Applies transformations to the given model.

        :param model: Target model.
        :param transformations: Transformations to apply to the model.
        :return: Target model after all transformations were applied.
        """
        for transformation in transformations:
            transformation.transformation_fn(model)
        return model



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import Dict, List, Optional, Tuple, Union

import torch.fx
from torch.ao.quantization.observer import HistogramObserver
from torch.ao.quantization.observer import PerChannelMinMaxObserver
from torch.ao.quantization.quantizer.quantizer import EdgeOrNode
from torch.ao.quantization.quantizer.quantizer import QuantizationAnnotation as TorchAOQuantizationAnnotation
from torch.ao.quantization.quantizer.quantizer import QuantizationSpec as TorchAOQuantizationSpec
from torch.ao.quantization.quantizer.quantizer import QuantizationSpecBase as TorchAOQuantizationSpecBase
from torch.ao.quantization.quantizer.quantizer import Quantizer as TorchAOQuantizer
from torch.ao.quantization.quantizer.quantizer import SharedQuantizationSpec as TorchAOSharedQuantizationSpec

import nncf
from nncf.common.graph.graph import NNCFGraph
from nncf.common.logging import nncf_logger
from nncf.common.quantization.quantizer_propagation.solver import QuantizerPropagationRule
from nncf.common.quantization.quantizer_setup import QuantizationPointBase
from nncf.common.quantization.quantizer_setup import SingleConfigQuantizerSetup
from nncf.common.quantization.structs import QuantizationPreset
from nncf.common.quantization.structs import QuantizationScheme
from nncf.experimental.torch.fx.nncf_graph_builder import GraphConverter
from nncf.experimental.torch.fx.node_utils import get_graph_node_by_name
from nncf.experimental.torch.fx.transformations import fold_constant_except_qdq
from nncf.parameters import ModelType
from nncf.parameters import QuantizationMode
from nncf.parameters import TargetDevice
from nncf.quantization.advanced_parameters import FP8QuantizationParameters
from nncf.quantization.advanced_parameters import OverflowFix
from nncf.quantization.advanced_parameters import QuantizationParameters
from nncf.quantization.algorithms.min_max.algorithm import MinMaxQuantization
from nncf.scopes import IgnoredScope
from nncf.torch.model_graph_manager import get_weight_tensor_port_ids

QUANT_ANNOTATION_KEY = "quantization_annotation"


class OpenVINOQuantizer(TorchAOQuantizer):
    """
    Implementation of the Torch AO quantizer which annotates models with quantization annotations
    optimally for the inference via OpenVINO.
    """

    def __init__(
        self,
        *,
        mode: Optional[QuantizationMode] = None,
        preset: Optional[QuantizationPreset] = None,
        target_device: TargetDevice = TargetDevice.ANY,
        model_type: Optional[ModelType] = None,
        ignored_scope: Optional[IgnoredScope] = None,
        overflow_fix: Optional[OverflowFix] = None,
        quantize_outputs: bool = False,
        activations_quantization_params: Optional[Union[QuantizationParameters, FP8QuantizationParameters]] = None,
        weights_quantization_params: Optional[Union[QuantizationParameters, FP8QuantizationParameters]] = None,
        quantizer_propagation_rule: QuantizerPropagationRule = QuantizerPropagationRule.MERGE_ALL_IN_ONE,
    ):
        """
        :param mode: Defines optimization mode for the algorithm. None by default.
        :param preset: A preset controls the quantization mode (symmetric and asymmetric).
            It can take the following values:
            - `performance`: Symmetric quantization of weights and activations.
            - `mixed`: Symmetric quantization of weights and asymmetric quantization of activations.
            Default value is None. In this case, `mixed` preset is used for `transformer`
            model type otherwise `performance`.
        :param target_device: A target device the specificity of which will be taken
            into account while compressing in order to obtain the best performance
            for this type of device, defaults to TargetDevice.ANY.
        :param model_type: Model type is needed to specify additional patterns
            in the model. Supported only `transformer` now.
        :param ignored_scope: An ignored scope that defined the list of model control
            flow graph nodes to be ignored during quantization.
        :param overflow_fix: This option controls whether to apply the overflow issue
            fix for the 8-bit quantization.
        :param quantize_outputs: Whether to insert additional quantizers right before
            each of the model outputs.
        :param activations_quantization_params: Quantization parameters for model
            activations.
        :param weights_quantization_params: Quantization parameters for model weights.
        :param quantizer_propagation_rule: The strategy to be used while propagating and merging quantizers.
        MERGE_ALL_IN_ONE by default.
        """
        self._min_max_algo = MinMaxQuantization(
            mode=mode,
            preset=preset,
            target_device=target_device,
            model_type=model_type,
            ignored_scope=ignored_scope,
            overflow_fix=overflow_fix,
            quantize_outputs=quantize_outputs,
            activations_quantization_params=activations_quantization_params,
            weights_quantization_params=weights_quantization_params,
            quantizer_propagation_rule=quantizer_propagation_rule,
        )

    def get_quantization_setup(self, model: torch.fx.GraphModule, nncf_graph: NNCFGraph) -> SingleConfigQuantizerSetup:
        self._min_max_algo._set_backend_entity(model)
        return self._min_max_algo.find_quantization_setup(model, nncf_graph)

    def annotate(self, model: torch.fx.GraphModule) -> torch.fx.GraphModule:
        nncf_graph = GraphConverter.create_nncf_graph(model)
        quantization_setup = self.get_quantization_setup(model, nncf_graph)

        graph = model.graph
        node_vs_torch_annotation = defaultdict(TorchAOQuantizationAnnotation)

        for qp in quantization_setup.quantization_points.values():
            edge_or_node, annotation = self._get_edge_or_node_and_annotation(
                graph, nncf_graph, qp, node_vs_torch_annotation
            )
            qspec = self._get_torch_ao_qspec_from_qp(qp)
            self._fill_torch_ao_annotation(edge_or_node, qspec, annotation)

        for quantizer_ids in quantization_setup.unified_scale_groups.values():

            root_quantizer_id = self._get_unified_scales_root_quantizer_id(
                nncf_graph, quantizer_ids, quantization_setup
            )
            root_qp = quantization_setup.quantization_points[root_quantizer_id]

            if any(root_qp.qconfig != quantization_setup.quantization_points[q_id].qconfig for q_id in quantizer_ids):
                qps = [quantization_setup.quantization_points[q_id] for q_id in quantizer_ids]
                raise nncf.InternalError(
                    "Different quantization configs are set to one unified scale group:"
                    f"{[(qp.insertion_point.__dict__, str(qp.qconfig)) for qp in qps]}"
                )

            root_target_node = get_graph_node_by_name(graph, root_qp.insertion_point.target_node_name)
            root_edge_or_node = self._get_edge_or_node(root_target_node, root_qp, nncf_graph)

            for quantizer_id in quantizer_ids:
                if quantizer_id == root_quantizer_id:
                    continue

                qspec = TorchAOSharedQuantizationSpec(root_edge_or_node)
                qp = quantization_setup.quantization_points[quantizer_id]
                edge_or_node, annotation = self._get_edge_or_node_and_annotation(
                    graph, nncf_graph, qp, node_vs_torch_annotation
                )
                self._fill_torch_ao_annotation(edge_or_node, qspec, annotation)

        for node, annotation in node_vs_torch_annotation.items():
            assert QUANT_ANNOTATION_KEY not in node.meta
            node.meta[QUANT_ANNOTATION_KEY] = annotation

    @staticmethod
    def _get_unified_scales_root_quantizer_id(
        nncf_graph: NNCFGraph, quantizer_ids: List[int], quantizer_setup: SingleConfigQuantizerSetup
    ) -> int:
        """
        Identifies the earliest quantizer node ID based on the corresponding `nncf_node.node_id`
        in the given NNCFGraph. This is required by the `_get_obs_or_fq_map` function.
        Refer to: https://github.com/pytorch/pytorch/blob/main/torch/ao/quantization/pt2e/prepare.py#L291

        :param nncf_graph: The NNCFGraph instance.
        :param quantizer_ids: The list of quantizer IDs to evaluate.
        :param quantizer_setup: The instance of SingleConfigQuantizerSetup.
        :return: The ID of the earliest quantizer node in terms of `nncf_node.node_id`.
        """
        nncf_node_quantizer_id = None
        root_quantizer_id = None
        for quantizer_id in quantizer_ids:
            target_node_name = quantizer_setup.quantization_points[quantizer_id].insertion_point.target_node_name
            nncf_node = nncf_graph.get_node_by_name(target_node_name)
            if nncf_node_quantizer_id is None or nncf_node.node_id < nncf_node_quantizer_id:
                root_quantizer_id = quantizer_id
                nncf_node_quantizer_id = nncf_node.node_id
        return root_quantizer_id

    @staticmethod
    def _get_edge_or_node_and_annotation(
        graph: torch.fx.Graph,
        nncf_graph: NNCFGraph,
        qp: QuantizationPointBase,
        node_vs_torch_annotation: Dict[torch.fx.Node, TorchAOQuantizationAnnotation],
    ) -> Tuple[EdgeOrNode, TorchAOQuantizationAnnotation]:
        """
        Retrieves the edge or node and its corresponding TorchAOQuantizationAnnotation based on the given graph,
        quantization point, and node-to-annotation mapping.

        :param graph: torch.fx.Graph instance.
        :param nncf_graph: NNCFGraph instance.
        :param qp: QuantizationPointBase instance.
        :param node_vs_torch_annotation: A dictionary mapping torch.fx.GraphNode objects to their respective
            TorchAOQuantizationAnnotations.
        :return: A tuple containing the EdgeOrNode and its associated TorchAOQuantizationAnnotation.
        """
        target_node = get_graph_node_by_name(graph, qp.insertion_point.target_node_name)
        annotation = node_vs_torch_annotation[target_node]
        edge_or_node = OpenVINOQuantizer._get_edge_or_node(target_node, qp, nncf_graph)
        return edge_or_node, annotation

    @staticmethod
    def _get_edge_or_node(target_node: torch.fx.Node, qp: QuantizationPointBase, nncf_graph: NNCFGraph) -> EdgeOrNode:
        """
        Returns the edge or node based on the given target node and quantization point.

        :param target_node: Target node instance.
        :param qp: QuantizationPointBase instance.
        :param graph: NNCFGraph instance.
        :return: The corresponding EdgeOrNode derived from the target node and quantization point.
        """
        ip = qp.insertion_point
        if qp.is_weight_quantization_point():
            nncf_node = nncf_graph.get_node_by_name(target_node.name)
            weights_ports_ids = get_weight_tensor_port_ids(nncf_node, nncf_graph)
            if len(weights_ports_ids) > 1:
                # TODO(dlyakhov): support quantization for nodes with several weights
                nncf_logger.warning(
                    f"Quantization of the weighted node {target_node.name}"
                    " is not yet supported by the OpenVINOQuantizer."
                    f" Only the weight on port ID {weights_ports_ids[0]} will be quantized."
                    f" Quantizable weights are located on ports: {weights_ports_ids}."
                )
            weight_node = target_node.all_input_nodes[weights_ports_ids[0]]
            return (weight_node, target_node)

        if ip.input_port_id is None:
            return target_node

        node = target_node.all_input_nodes[ip.input_port_id]
        return (node, target_node)

    @staticmethod
    def _fill_torch_ao_annotation(
        edge_or_node: EdgeOrNode,
        qspec: TorchAOQuantizationSpecBase,
        annotation_to_update: TorchAOQuantizationAnnotation,
    ) -> None:
        """
        Helper method to update the annotation_to_update based on the specified edge_or_node and qspec.

        :param edge_or_node: The target EdgeOrNode to be used for the update.
        :param qspec: An instance of TorchAOQuantizationSpecBase representing the quantization specification to apply.
        :param annotation_to_update: The annotation to update based on the edge_or_node and qspec.
        """
        if isinstance(edge_or_node, torch.fx.Node):
            annotation_to_update.output_qspec = qspec
        else:
            annotation_to_update.input_qspec_map[edge_or_node[0]] = qspec

    @staticmethod
    def _get_torch_ao_qspec_from_qp(qp: QuantizationPointBase) -> TorchAOQuantizationSpec:
        """
        Retrieves the quantization configuration from the given quantization point and
        converts it into a TorchAOQuantizationSpec.

        :param qp: An instance of QuantizationPointBase.
        :return: A TorchAOQuantizationSpec retrieved and converted from the quantization point.
        """
        # Eps value is copied from nncf/torch/quantization/layers.py
        extra_args = {"eps": 1e-16}
        qconfig = qp.qconfig
        is_weight = qp.is_weight_quantization_point()

        if qconfig.per_channel:
            torch_qscheme = (
                torch.per_channel_symmetric
                if qconfig.mode is QuantizationScheme.SYMMETRIC
                else torch.per_channel_affine
            )
        else:
            torch_qscheme = (
                torch.per_tensor_symmetric if qconfig.mode is QuantizationScheme.SYMMETRIC else torch.per_tensor_affine
            )
        if is_weight:
            observer = PerChannelMinMaxObserver
            quant_min = -128
            quant_max = 127
            dtype = torch.int8
            channel_axis = 0
        else:
            observer = (
                HistogramObserver
                if torch_qscheme in [torch.per_tensor_symmetric, torch.per_tensor_affine]
                else PerChannelMinMaxObserver
            )
            quant_min = 0
            quant_max = 255
            dtype = torch.int8 if qconfig.signedness_to_force else torch.uint8
            channel_axis = 1  # channel dim for activations
        return TorchAOQuantizationSpec(
            dtype=dtype,
            observer_or_fake_quant_ctr=observer.with_args(**extra_args),
            quant_min=quant_min,
            quant_max=quant_max,
            qscheme=torch_qscheme,
            ch_axis=channel_axis,
            is_dynamic=False,
        )

    def validate(self, model: torch.fx.GraphModule) -> None:
        pass

    def transform_for_annotation(self, model: torch.fx.GraphModule) -> torch.fx.GraphModule:
        fold_constant_except_qdq(model)
        return model



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.tensor.definitions import TensorBackend as TensorBackend
from nncf.tensor.definitions import TensorDataType as TensorDataType
from nncf.tensor.definitions import TensorDeviceType as TensorDeviceType
from nncf.tensor.tensor import Tensor as Tensor
from nncf.tensor.tensor import unwrap_tensor_data as unwrap_tensor_data



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional, Set, Tuple

import torch.fx

import nncf
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.transformations.commands import TargetType
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.experimental.torch.fx.commands import FXApplyTransformationCommand
from nncf.experimental.torch.fx.model_utils import get_target_point
from nncf.experimental.torch.fx.model_utils import remove_fq_from_inputs
from nncf.experimental.torch.fx.node_utils import get_bias_value
from nncf.experimental.torch.fx.node_utils import get_graph_node_by_name
from nncf.experimental.torch.fx.node_utils import is_node_with_bias
from nncf.experimental.torch.fx.transformations import constant_update_transformation_builder
from nncf.experimental.torch.fx.transformations import output_insertion_transformation_builder
from nncf.quantization.algorithms.bias_correction.backend import BiasCorrectionAlgoBackend
from nncf.tensor import Tensor
from nncf.torch.graph.transformations.commands import PTModelExtractionCommand
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.model_graph_manager import is_quantized_weights
from nncf.torch.tensor_statistics.collectors import get_mean_statistic_collector
from nncf.torch.tensor_statistics.collectors import get_raw_stat_collector


class FXBiasCorrectionAlgoBackend(BiasCorrectionAlgoBackend):

    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> PTTargetPoint:
        return get_target_point(target_type, target_node_name, port_id)

    @staticmethod
    def create_bias_correction_command(
        node: NNCFNode, bias_value: Tensor, nncf_graph: NNCFGraph
    ) -> FXApplyTransformationCommand:
        return FXApplyTransformationCommand(
            constant_update_transformation_builder(node, bias_value.data, input_port_id=2)
        )

    @staticmethod
    def model_extraction_command(
        input_ids: Set[Tuple[str, int]], output_ids: Set[Tuple[str, int]]
    ) -> PTModelExtractionCommand:
        return PTModelExtractionCommand([inp_id[0] for inp_id in input_ids], [out_id[0] for out_id in output_ids])

    @staticmethod
    def output_insertion_command(nncf_graph: NNCFGraph, target_point: PTTargetPoint) -> FXApplyTransformationCommand:
        return FXApplyTransformationCommand(output_insertion_transformation_builder(target_point))

    @staticmethod
    def mean_statistic_collector(
        channel_axis: int,
        inplace: bool,
        num_samples: Optional[int] = None,
        window_size: Optional[int] = None,
    ) -> TensorCollector:
        return get_mean_statistic_collector(num_samples, channel_axis, window_size)

    @staticmethod
    def raw_statistic_collector(num_samples: Optional[int] = None) -> TensorCollector:
        return get_raw_stat_collector(num_samples)

    @staticmethod
    def process_model_output(raw_data: Dict, output_name: int) -> Tensor:
        return Tensor(raw_data[output_name])

    @staticmethod
    def get_activation_port_id(node: NNCFNode, nncf_graph: NNCFGraph) -> int:
        return 0

    @staticmethod
    def get_bias_value(node: NNCFNode, model: torch.fx.GraphModule, nncf_graph: NNCFGraph) -> Tensor:
        return Tensor(get_bias_value(node, nncf_graph, model))

    @staticmethod
    def get_input_name(model: torch.fx.GraphModule, node_name: str, input_port_id: int) -> str:
        graph_node = get_graph_node_by_name(model.graph, node_name)
        return graph_node.all_input_nodes[input_port_id].name

    @staticmethod
    def get_output_name(model: torch.fx.GraphModule, node_name: str, output_port_id: int) -> int:
        graph_node = get_graph_node_by_name(model.graph, node_name)
        if graph_node.op == "output":
            # Original node output is kept as the first
            # output tensor, thus returns 0.
            return 0
        nodes = list(graph_node.users)
        while nodes:
            node = nodes.pop()
            if node.op == "call_function" and node.target == torch.ops.aten.clone.default:
                nodes = list(node.users)
                graph_node = node
            elif node.op == "output":
                return node.all_input_nodes.index(graph_node)

        raise nncf.InternalError(f"Node with name {node_name} expected to have an output," " no outputs were found.")

    @staticmethod
    def is_quantized_weights(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return is_quantized_weights(node, nncf_graph)

    @staticmethod
    def is_node_with_bias(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return is_node_with_bias(node, nncf_graph)

    @staticmethod
    def remove_fq_from_inputs(model: torch.fx.GraphModule, nncf_graph: NNCFGraph) -> torch.fx.GraphModule:
        return remove_fq_from_inputs(model, nncf_graph)

    @staticmethod
    def get_port_id(target_point: PTTargetPoint) -> int:
        if target_point.target_type == TargetType.OPERATOR_POST_HOOK:
            # Return 0 as default value for post hook port id.
            return 0
        return target_point.input_port_id



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from nncf.experimental.common.tensor_statistics.collectors import AbsMaxReducer
from nncf.experimental.common.tensor_statistics.collectors import AbsQuantileReducer
from nncf.experimental.common.tensor_statistics.collectors import BatchMeanReducer
from nncf.experimental.common.tensor_statistics.collectors import InplaceInsertionFNType
from nncf.experimental.common.tensor_statistics.collectors import MaxReducer
from nncf.experimental.common.tensor_statistics.collectors import MaxVarianceReducer
from nncf.experimental.common.tensor_statistics.collectors import MeanAbsMaxReducer
from nncf.experimental.common.tensor_statistics.collectors import MeanAggregator
from nncf.experimental.common.tensor_statistics.collectors import MeanPerChReducer
from nncf.experimental.common.tensor_statistics.collectors import MeanReducer
from nncf.experimental.common.tensor_statistics.collectors import MeanVarianceReducer
from nncf.experimental.common.tensor_statistics.collectors import MinReducer
from nncf.experimental.common.tensor_statistics.collectors import NoopAggregator
from nncf.experimental.common.tensor_statistics.collectors import QuantileReducer
from nncf.experimental.common.tensor_statistics.collectors import RawReducer
from nncf.experimental.common.tensor_statistics.collectors import ShapeAggregator
from nncf.experimental.common.tensor_statistics.collectors import ShapeReducer
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.experimental.common.tensor_statistics.statistics import MeanTensorStatistic
from nncf.experimental.common.tensor_statistics.statistics import RawTensorStatistic
from nncf.openvino.graph.node_utils import get_inplace_batch_mean_op
from nncf.openvino.graph.node_utils import get_inplace_max_op
from nncf.openvino.graph.node_utils import get_inplace_max_var_op
from nncf.openvino.graph.node_utils import get_inplace_mean_max_op
from nncf.openvino.graph.node_utils import get_inplace_mean_op
from nncf.openvino.graph.node_utils import get_inplace_mean_per_ch
from nncf.openvino.graph.node_utils import get_inplace_mean_var_op
from nncf.openvino.graph.node_utils import get_inplace_min_op
from nncf.openvino.graph.node_utils import get_inplace_shape_op
from nncf.quantization.advanced_parameters import StatisticsType


class OVMinReducer(MinReducer):

    def get_inplace_fn(self):
        return get_inplace_min_op(self._reduction_axes)


class OVMaxReducer(MaxReducer):

    def get_inplace_fn(self):
        return get_inplace_max_op(self._reduction_axes, False)


class OVAbsMaxReducer(AbsMaxReducer):

    def get_inplace_fn(self):
        return get_inplace_max_op(self._reduction_axes, True)


class OVMeanReducer(MeanReducer):

    def get_inplace_fn(self):
        return get_inplace_mean_op(self._reduction_axes)


class OVMeanVarianceReducer(MeanVarianceReducer):
    def get_inplace_fn(self):
        return get_inplace_mean_var_op(self._reduction_axes)


class OVMaxVarianceReducer(MaxVarianceReducer):
    def get_inplace_fn(self):
        return get_inplace_max_var_op(self._reduction_axes)


class OVMeanAbsMaxReducer(MeanAbsMaxReducer):
    def get_inplace_fn(self):
        return get_inplace_mean_max_op(self._reduction_axes, True)


class OVShapeReducer(ShapeReducer):
    def get_inplace_fn(self) -> Optional[InplaceInsertionFNType]:
        return get_inplace_shape_op()


class OVBatchMeanReducer(BatchMeanReducer):

    def get_inplace_fn(self):
        return get_inplace_batch_mean_op()


class OVMeanPerChanelReducer(MeanPerChReducer):

    def get_inplace_fn(self):
        return get_inplace_mean_per_ch(self._channel_axis)


class OVQuantileReducer(QuantileReducer):
    def get_inplace_fn(self) -> Optional[InplaceInsertionFNType]:
        return None


class OVAbsQuantileReducer(AbsQuantileReducer):

    def get_inplace_fn(self) -> Optional[InplaceInsertionFNType]:
        return None


def get_mean_statistic_collector(
    num_samples: int, channel_axis: int, window_size: Optional[int] = None, inplace: bool = True
) -> TensorCollector:
    """
    Mean statistic collector builder.

    :param num_samples: Maximum number of samples to collect.
    :param channel_axis: Channel axis to use during reduction phase.
    :param window_size: Number of samples from the end of the list of collected samples to aggregate.
        Aggregates all available collected statistics in case parameter is None.
    :param inplace: Whether the mean reducer should be calculated inplace or out of place.
    :return: Mean statistic collector.
    """
    if channel_axis == 0:
        reducer = OVBatchMeanReducer(inplace)
    else:
        reducer = OVMeanPerChanelReducer(channel_axis=channel_axis, inplace=inplace)
    raw_reducer = RawReducer()

    kwargs = {
        "num_samples": num_samples,
        "window_size": window_size,
    }
    aggregate_mean = MeanAggregator(**kwargs)
    aggregate_shape = ShapeAggregator()

    collector = TensorCollector(MeanTensorStatistic)
    collector.register_statistic_branch(MeanTensorStatistic.MEAN_STAT, reducer, aggregate_mean)
    collector.register_statistic_branch(MeanTensorStatistic.SHAPE_STAT, raw_reducer, aggregate_shape)
    return collector


def get_raw_stat_collector(num_samples: Optional[int] = None) -> TensorCollector:
    reducer = RawReducer()
    aggregator = NoopAggregator(num_samples)

    collector = TensorCollector(RawTensorStatistic)
    collector.register_statistic_branch(RawTensorStatistic.VALUES_STATS, reducer, aggregator)
    return collector


OV_REDUCERS_MAP = {
    StatisticsType.MIN: OVMinReducer,
    StatisticsType.MAX: OVMaxReducer,
    StatisticsType.ABS_MAX: OVAbsMaxReducer,
    StatisticsType.MEAN: OVMeanReducer,
    StatisticsType.QUANTILE: OVQuantileReducer,
    StatisticsType.ABS_QUANTILE: OVAbsQuantileReducer,
}
